# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from typing import List, Union

import numpy as np

from .... import opcodes as OperandDef
from ....core import get_output_types, recursive_tile
from ....core.context import get_context, Context
from ....dataframe.core import DATAFRAME_TYPE
from ....serialization.serializables import KeyField, Float64Field, ListField, BoolField
from ....tensor.core import TENSOR_TYPE, TENSOR_CHUNK_TYPE
from ....tensor import tensor as astensor
from ....typing import TileableType, ChunkType
from ....utils import has_unknown_shape, ensure_own_data, build_fetch
from ...operands import LearnOperand, LearnOperandMixin
from ...utils import convert_to_tensor_or_dataframe, concat_chunks


class ToDMatrix(LearnOperand, LearnOperandMixin):
    _op_type_ = OperandDef.TO_DMATRIX

    data = KeyField("data")
    label = KeyField("label")
    missing = Float64Field("missing")
    weight = KeyField("weight")
    base_margin = KeyField("base_margin")
    feature_names = ListField("feature_names")
    feature_types = ListField("feature_types")
    # if to collocate the data, label and weight
    _collocate = BoolField("collocate", default=False)

    @property
    def output_limit(self):
        if self._collocate:
            return (
                1
                + (self.label is not None)
                + (self.weight is not None)
                + (self.base_margin is not None)
            )
        return 1

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if self.data is not None:
            self.data = self._inputs[0]
        has_label = self.label is not None
        if has_label:
            self.label = self._inputs[1]
        if self.weight is not None:
            i = 1 if not has_label else 2
            self.weight = self._inputs[i]
        if self.base_margin is not None:
            self.base_margin = self._inputs[-1]

    @staticmethod
    def _get_kw(obj):
        if isinstance(obj, TENSOR_TYPE + TENSOR_CHUNK_TYPE):
            return {"shape": obj.shape, "dtype": obj.dtype, "order": obj.order}
        else:
            return {
                "shape": obj.shape,
                "dtypes": obj.dtypes,
                "index_value": obj.index_value,
                "columns_value": obj.columns_value,
            }

    def __call__(self):
        inputs = [self.data]
        kw = self._get_kw(self.data)
        if self.label is not None:
            inputs.append(self.label)
        if self.weight is not None:
            inputs.append(self.weight)
        if self.base_margin is not None:
            inputs.append(self.base_margin)

        return self.new_tileable(inputs, **kw)

    @classmethod
    def _get_collocated(
        cls,
        op: "ToDMatrix",
        data: TileableType,
        label: TileableType,
        weight: TileableType,
        base_margin: TileableType,
    ) -> List[TileableType]:
        types = ["data", "label", "weight", "base_margin"]
        nsplit = data.nsplits[0]
        out_chunkss = [[] for _ in op.inputs]
        for i in range(len(nsplit)):
            data_chunk = data.cix[i, 0]
            inps = [data_chunk]
            kws = []
            chunk_op = op.copy().reset_key()
            chunk_op._collocate = True
            chunk_op.data = data_chunk
            output_types = [get_output_types(data)[0]]
            data_kw = cls._get_kw(data_chunk)
            data_kw["index"] = data_chunk.index
            kws.append(data_kw)
            for type_name, inp in zip(types[1:], [label, weight, base_margin]):
                if inp is None:
                    continue
                inp_chunk = inp.cix[
                    i,
                ]
                setattr(chunk_op, type_name, inp_chunk)
                inps.append(inp_chunk)
                kw = cls._get_kw(inp_chunk)
                kw["index"] = inp_chunk.index
                kw["type"] = type_name
                kws.append(kw)
                output_types.append(get_output_types(inp)[0])
            chunk_op.output_types = output_types
            out_chunks = chunk_op.new_chunks(inps, kws=kws)
            for i, out_chunk in enumerate(out_chunks):
                out_chunkss[i].append(out_chunk)

        new_op = op.copy()
        new_op._collocate = True
        outs = [data, label, weight, base_margin]
        params = [out.params.copy() for out in outs if out is not None]
        output_types = []
        j = 0
        for i, out in enumerate(outs):
            if out is None:
                continue
            params[j]["nsplits"] = out.nsplits
            params[j]["chunks"] = out_chunkss[j]
            params[j]["type"] = types[i]
            output_types.append(get_output_types(out)[0])
            j += 1
        new_op.output_types = output_types
        return new_op.new_tileables(op.inputs, kws=params)

    @staticmethod
    def _order_chunk_index(chunks: List[ChunkType]):
        ndim = chunks[0].ndim
        for i, c in enumerate(chunks):
            if ndim == 2:
                c._index = (i, 0)
            else:
                c._index = (i,)
        return chunks

    @classmethod
    def tile(cls, op: "MarsDMatrix"):
        data, label, weight, base_margin = op.data, op.label, op.weight, op.base_margin

        if has_unknown_shape(data):
            yield
        if data.chunk_shape[1] > 1:
            # make sure data's second dimension has only 1 chunk
            data = yield from recursive_tile(data.rechunk({1: data.shape[1]}))
        nsplit = data.nsplits[0]
        # rechunk label
        if label is not None:
            label = yield from recursive_tile(label.rechunk({0: nsplit}))
        # rechunk weight
        if weight is not None:
            weight = yield from recursive_tile(weight.rechunk({0: nsplit}))
        # rechunk base_margin
        if base_margin is not None:
            base_margin = yield from recursive_tile(base_margin.rechunk({0: nsplit}))

        collocated = cls._get_collocated(op, data, label, weight, base_margin)
        collocated_chunks = list(
            itertools.chain.from_iterable(c.chunks for c in collocated)
        )
        yield collocated_chunks + collocated

        data = build_fetch(collocated[0])
        has_label = False
        if label is not None:
            has_label = True
            label = build_fetch(collocated[1])
        i_weight = -1
        if weight is not None:
            i_weight = 1 if not has_label else 2
            weight = build_fetch(collocated[i_weight])
        if base_margin is not None:
            base_margin = build_fetch(collocated[-1])

        ctx = get_context()

        # for distributed, we should concat the chunks
        # which allocated on the same worker into one
        data_chunk_metas = ctx.get_chunks_meta(
            [c.key for c in data.chunks], fields=["bands"]
        )
        data_chunk_workers = [m["bands"][0][0] for m in data_chunk_metas]
        worker_to_chunks = dict()
        for i, worker in enumerate(data_chunk_workers):
            size = 1 + sum(it is not None for it in [label, weight, base_margin])
            if worker not in worker_to_chunks:
                worker_to_chunks[worker] = [[] for _ in range(size)]
            worker_to_chunks[worker][0].append(data.chunks[i])
            if label is not None:
                worker_to_chunks[worker][1].append(label.chunks[i])
            if weight is not None:
                worker_to_chunks[worker][i_weight].append(weight.chunks[i])
            if base_margin is not None:
                worker_to_chunks[worker][-1].append(base_margin.chunks[i])
        ind = itertools.count(0)
        out_chunks = []
        for worker, chunks in worker_to_chunks.items():
            data_chunk = concat_chunks(cls._order_chunk_index(chunks[0]))
            inps = [data_chunk]
            label_chunk = None
            if label is not None:
                label_chunk = concat_chunks(cls._order_chunk_index(chunks[1]))
                inps.append(label_chunk)
            weight_chunk = None
            if weight is not None:
                weight_chunk = concat_chunks(cls._order_chunk_index(chunks[i_weight]))
                inps.append(weight_chunk)
            base_margin_chunk = None
            if base_margin is not None:
                base_margin_chunk = concat_chunks(cls._order_chunk_index(chunks[-1]))
                inps.append(base_margin_chunk)
            chunk_op = ToDMatrix(
                data=data_chunk,
                label=label_chunk,
                missing=op.missing,
                weight=weight_chunk,
                base_margin=base_margin_chunk,
                feature_names=op.feature_names,
                feature_types=op.feature_types,
                _output_types=op.output_types,
            )
            kws = data_chunk.params
            kws["index"] = (next(ind), 0)
            out_chunks.append(chunk_op.new_chunk(inps, **kws))
        nsplits = (tuple(c.shape[0] for c in out_chunks), (out_chunks[0].shape[1],))

        new_op = op.copy()
        kw = op.outputs[0].params
        kw["chunks"] = out_chunks
        kw["nsplits"] = nsplits
        return new_op.new_tileables(op.inputs, kws=[kw])

    @staticmethod
    def get_xgb_dmatrix(tup, nthread: int = -1):
        from xgboost import DMatrix

        data, label, weight, base_margin, missing, feature_names, feature_types = tup
        data = data.spmatrix if hasattr(data, "spmatrix") else data
        return DMatrix(
            ensure_own_data(data),
            label=ensure_own_data(label),
            missing=missing,
            weight=ensure_own_data(weight),
            base_margin=base_margin,
            feature_names=feature_names,
            feature_types=feature_types,
            nthread=nthread,
        )

    @staticmethod
    def _from_ctx_if_not_none(ctx, chunk):
        if chunk is None:
            return chunk
        return ctx[chunk.key]

    @classmethod
    def execute(cls, ctx: Union[dict, Context], op: "ToDMatrix"):
        if op._collocate:
            outs = op.outputs
            ctx[outs[0].key] = ctx[op.inputs[0].key]
            has_label = False
            if op.label is not None:
                has_label = True
                ctx[outs[1].key] = ctx[op.inputs[1].key]
            if op.weight is not None:
                i_weight = 1 if not has_label else 2
                ctx[outs[i_weight].key] = ctx[op.inputs[i_weight].key]
            if op.base_margin is not None:
                ctx[outs[-1].key] = ctx[op.inputs[-1].key]
        else:
            out = op.outputs[0]
            data = cls._from_ctx_if_not_none(ctx, op.data)
            if data is None:
                data = np.empty((0, out.shape[1]))
            ctx[out.key] = (
                data,
                cls._from_ctx_if_not_none(ctx, op.label),
                cls._from_ctx_if_not_none(ctx, op.weight),
                cls._from_ctx_if_not_none(ctx, op.base_margin),
                op.missing,
                op.feature_names,
                op.feature_types,
            )


def check_data(data):
    data = convert_to_tensor_or_dataframe(data)
    if data.ndim != 2:
        raise ValueError(f"Expecting 2-d data, got: {data.ndim}-d")

    return data


def check_array_like(y: TileableType, name: str) -> TileableType:
    if y is None:
        return
    y = convert_to_tensor_or_dataframe(y)
    if isinstance(y, DATAFRAME_TYPE):
        y = y.iloc[:, 0]
    y = astensor(y)
    if y.ndim != 1:
        raise ValueError(f"Expecting 1-d {name}, got: {y.ndim}-d")
    return y


def to_dmatrix(
    data,
    label=None,
    missing=None,
    weight=None,
    base_margin=None,
    feature_names=None,
    feature_types=None,
):
    data = check_data(data)
    label = check_array_like(label, "label")
    weight = check_array_like(weight, "weight")
    base_margin = check_array_like(base_margin, "base_margin")

    # If not multiple outputs, try to collect the chunks on same worker into one
    # to feed the data into XGBoost for training.
    op = ToDMatrix(
        data=data,
        label=label,
        missing=missing,
        weight=weight,
        base_margin=base_margin,
        feature_names=feature_names,
        feature_types=feature_types,
        gpu=data.op.gpu,
        _output_types=get_output_types(data),
    )
    return op()


MarsDMatrix = to_dmatrix



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import AnyField
from ..utils import gen_random_seeds
from .core import TensorRandomOperandMixin, handle_array, TensorDistribution


class TensorBinomial(TensorDistribution, TensorRandomOperandMixin):
    _input_fields_ = ["n", "p"]
    _op_type_ = OperandDef.RAND_BINOMIAL

    _fields_ = "n", "p", "size"
    n = AnyField("n")
    p = AnyField("p")
    _func_name = "binomial"

    def __call__(self, n, p, chunk_size=None):
        return self.new_tensor([n, p], None, raw_chunk_size=chunk_size)


def binomial(random_state, n, p, size=None, chunk_size=None, gpu=None, dtype=None):
    r"""
    Draw samples from a binomial distribution.

    Samples are drawn from a binomial distribution with specified
    parameters, n trials and p probability of success where
    n an integer >= 0 and p is in the interval [0,1]. (n may be
    input as a float, but it is truncated to an integer in use)

    Parameters
    ----------
    n : int or array_like of ints
        Parameter of the distribution, >= 0. Floats are also accepted,
        but they will be truncated to integers.
    p : float or array_like of floats
        Parameter of the distribution, >= 0 and <=1.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``n`` and ``p`` are both scalars.
        Otherwise, ``mt.broadcast(n, p).size`` samples are drawn.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default
    dtype : data-type, optional
      Data-type of the returned tensor.

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized binomial distribution, where
        each sample is equal to the number of successes over the n trials.

    See Also
    --------
    scipy.stats.binom : probability density function, distribution or
        cumulative density function, etc.

    Notes
    -----
    The probability density for the binomial distribution is

    .. math:: P(N) = \binom{n}{N}p^N(1-p)^{n-N},

    where :math:`n` is the number of trials, :math:`p` is the probability
    of success, and :math:`N` is the number of successes.

    When estimating the standard error of a proportion in a population by
    using a random sample, the normal distribution works well unless the
    product p*n <=5, where p = population proportion estimate, and n =
    number of samples, in which case the binomial distribution is used
    instead. For example, a sample of 15 people shows 4 who are left
    handed, and 11 who are right handed. Then p = 4/15 = 27%. 0.27*15 = 4,
    so the binomial distribution should be used in this case.

    References
    ----------
    .. [1] Dalgaard, Peter, "Introductory Statistics with R",
           Springer-Verlag, 2002.
    .. [2] Glantz, Stanton A. "Primer of Biostatistics.", McGraw-Hill,
           Fifth Edition, 2002.
    .. [3] Lentner, Marvin, "Elementary Applied Statistics", Bogden
           and Quigley, 1972.
    .. [4] Weisstein, Eric W. "Binomial Distribution." From MathWorld--A
           Wolfram Web Resource.
           http://mathworld.wolfram.com/BinomialDistribution.html
    .. [5] Wikipedia, "Binomial distribution",
           http://en.wikipedia.org/wiki/Binomial_distribution

    Examples
    --------
    Draw samples from the distribution:

    >>> import mars.tensor as mt

    >>> n, p = 10, .5  # number of trials, probability of each trial
    >>> s = mt.random.binomial(n, p, 1000).execute()
    # result of flipping a coin 10 times, tested 1000 times.

    A real world example. A company drills 9 wild-cat oil exploration
    wells, each with an estimated probability of success of 0.1. All nine
    wells fail. What is the probability of that happening?

    Let's do 20,000 trials of the model, and count the number that
    generate zero positive results.

    >>> (mt.sum(mt.random.binomial(9, 0.1, 20000) == 0)/20000.).execute()
    # answer = 0.38885, or 38%.
    """
    if dtype is None:
        dtype = (
            np.random.RandomState()
            .binomial(handle_array(n), handle_array(p), size=(0,))
            .dtype
        )
    size = random_state._handle_size(size)
    seed = gen_random_seeds(1, random_state.to_numpy())[0]
    op = TensorBinomial(seed=seed, size=size, gpu=gpu, dtype=dtype)
    return op(n, p, chunk_size=chunk_size)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Dict, Tuple, Iterator
from urllib import parse

from ...utils import implements, lazy_import
from ._oss_lib import common as oc
from ._oss_lib.glob import glob
from ._oss_lib.handle import OSSIOBase
from .base import FileSystem, path_type

oss2 = lazy_import("oss2", placeholder=True)

_oss_time_out = 10


class OSSFileSystem(FileSystem):
    _instance = None

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = OSSFileSystem()
        return cls._instance

    @implements(FileSystem.cat)
    def cat(self, path: path_type):
        raise NotImplementedError

    @implements(FileSystem.ls)
    def ls(self, path: path_type) -> List[path_type]:
        file_list = []
        file_entry = oc.OSSFileEntry(path)
        if not file_entry.is_dir():
            raise OSError("ls for file is not supported")
        else:
            bucket, key, access_key_id, access_key_secret, end_point = oc.parse_osspath(
                path
            )
            oss_bucket = oss2.Bucket(
                auth=oss2.Auth(
                    access_key_id=access_key_id, access_key_secret=access_key_secret
                ),
                endpoint=end_point,
                bucket_name=bucket,
                connect_timeout=_oss_time_out,
            )
            for obj in oss2.ObjectIteratorV2(oss_bucket, prefix=key):
                if obj.key.endswith("/"):
                    continue
                obj_path = rf"oss://{bucket}/{obj.key}"
                file_list.append(
                    build_oss_path(
                        obj_path, access_key_id, access_key_secret, end_point
                    )
                )
        return file_list

    @implements(FileSystem.delete)
    def delete(self, path: path_type, recursive: bool = False):
        raise NotImplementedError

    @implements(FileSystem.rename)
    def rename(self, path: path_type, new_path: path_type):
        raise NotImplementedError

    @implements(FileSystem.stat)
    def stat(self, path: path_type) -> Dict:
        ofe = oc.OSSFileEntry(path)
        return ofe.stat()

    @implements(FileSystem.mkdir)
    def mkdir(self, path: path_type, create_parents: bool = True):
        raise NotImplementedError

    @implements(FileSystem.isdir)
    def isdir(self, path: path_type) -> bool:
        file_entry = oc.OSSFileEntry(path)
        return file_entry.is_dir()

    @implements(FileSystem.isfile)
    def isfile(self, path: path_type) -> bool:
        file_entry = oc.OSSFileEntry(path)
        return file_entry.is_file()

    @implements(FileSystem._isfilestore)
    def _isfilestore(self) -> bool:
        raise NotImplementedError

    @implements(FileSystem.exists)
    def exists(self, path: path_type):
        return oc.oss_exists(path)

    @implements(FileSystem.open)
    def open(self, path: path_type, mode: str = "rb") -> OSSIOBase:
        file_handle = OSSIOBase(path, mode)
        return file_handle

    @implements(FileSystem.walk)
    def walk(self, path: path_type) -> Iterator[Tuple[str, List[str], List[str]]]:
        raise NotImplementedError

    @implements(FileSystem.glob)
    def glob(self, path: path_type, recursive: bool = False) -> List[path_type]:
        return glob(path, recursive=recursive)


def build_oss_path(path: path_type, access_key_id, access_key_secret, end_point):
    """
    Returns a path with oss info.
    Used to register the access_key_id, access_key_secret and
    endpoint of OSS. The access_key_id and endpoint are put
    into the url with url-safe-base64 encoding.

    Parameters
    ----------
    path : path_type
        The original oss url.

    access_key_id : str
        The access key id of oss.

    access_key_secret : str
        The access key secret of oss.

    end_point : str
        The endpoint of oss.

    Returns
    -------
    path_type
        Path include the encoded access key id, end point and
        access key secret of oss.
    """
    if isinstance(path, (list, tuple)):
        path = path[0]
    param_dict = {"access_key_id": access_key_id, "end_point": end_point}
    id_endpoint = oc.dict_to_url(param_dict)
    password = access_key_secret
    parse_result = parse.urlparse(path)
    new_path = (
        f"{parse_result.scheme}://{id_endpoint}:{password}"
        f"@{parse_result.netloc}{parse_result.path}"
    )
    return new_path



# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import subprocess
import sys
import uuid
from collections import namedtuple
from ctypes import (
    c_char,
    c_char_p,
    c_int,
    c_uint,
    c_ulonglong,
    byref,
    create_string_buffer,
    Structure,
    POINTER,
    CDLL,
)

from ..utils import parse_readable_size

logger = logging.getLogger(__name__)

# Some constants taken from cuda.h
CUDA_SUCCESS = 0
CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT = 16
CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_MULTIPROCESSOR = 39
CU_DEVICE_ATTRIBUTE_CLOCK_RATE = 13
CU_DEVICE_ATTRIBUTE_PCI_BUS_ID = 33
CU_DEVICE_ATTRIBUTE_PCI_DEVICE_ID = 34
CU_DEVICE_ATTRIBUTE_MEMORY_CLOCK_RATE = 36

CU_NO_CUDA_CAPABLE_DEVICE_DETECTED = 100

# nvml constants
NVML_SUCCESS = 0
NVML_TEMPERATURE_GPU = 0

NVML_DRIVER_NOT_LOADED = 9


class _CUuuid_t(Structure):
    _fields_ = [("bytes", c_char * 16)]


class _nvmlUtilization_t(Structure):
    _fields_ = [
        ("gpu", c_uint),
        ("memory", c_uint),
    ]


class _struct_nvmlDevice_t(Structure):
    pass  # opaque handle


_nvmlDevice_t = POINTER(_struct_nvmlDevice_t)


class _nvmlBAR1Memory_t(Structure):
    _fields_ = [
        ("total", c_ulonglong),
        ("free", c_ulonglong),
        ("used", c_ulonglong),
    ]


_is_windows: bool = sys.platform.startswith("win")
_is_wsl: bool = "WSL_DISTRO_NAME" in os.environ


def _load_nv_library(*libnames):
    for lib in libnames:
        try:
            return CDLL(lib)
        except OSError:
            continue


_cuda_lib = _nvml_lib = None

_cu_device_info = namedtuple(
    "_cu_device_info", "index uuid name multiprocessors cuda_cores threads"
)
_nvml_driver_info = namedtuple("_nvml_driver_info", "driver_version cuda_version")
_nvml_device_status = namedtuple(
    "_nvml_device_status",
    "gpu_util mem_util temperature fb_total_mem fb_used_mem fb_free_mem",
)

_init_pid = None
_gpu_count = None
_driver_info = None
_device_infos = dict()

_no_device_warned = False


class NVError(Exception):
    def __init__(self, msg, *args, errno=None):
        self._errno = errno
        super().__init__(msg or "Unknown error", *args)

    def __str__(self):
        return f"({self._errno}) {super().__str__()}"

    @property
    def errno(self):
        return self._errno

    @property
    def message(self):
        return super().__str__()


class NVDeviceAPIError(NVError):
    pass


class NVMLAPIError(NVError):
    pass


def _cu_check_error(result):
    if result != CUDA_SUCCESS:
        _error_str = c_char_p()
        _cuda_lib.cuGetErrorString(result, byref(_error_str))
        err_value = _error_str.value.decode() if _error_str.value is not None else None
        raise NVDeviceAPIError(err_value, errno=result)


_nvmlErrorString = None


def _nvml_check_error(result):
    global _nvmlErrorString
    if _nvmlErrorString is None:
        _nvmlErrorString = _nvml_lib.nvmlErrorString
        _nvmlErrorString.restype = c_char_p

    if result != NVML_SUCCESS:
        _error_str = _nvmlErrorString(result)
        raise NVMLAPIError(_error_str.decode(), errno=result)


_cu_process_var_to_cores = {
    (1, 0): 8,
    (1, 1): 8,
    (1, 2): 8,
    (1, 3): 8,
    (2, 0): 32,
    (2, 1): 48,
}


def _cu_get_processor_cores(major, minor):
    return _cu_process_var_to_cores.get((major, minor), 192)


def _init_cp():
    global _cuda_lib, _no_device_warned
    if _init_pid == os.getpid():
        return

    libcuda_paths = ["libcuda.so", "libcuda.dylib", "cuda.dll", "nvcuda.dll"]
    if _is_wsl:
        libcuda_paths = ["/usr/lib/wsl/lib/libcuda.so"] + libcuda_paths
    _cuda_lib = _load_nv_library(*libcuda_paths)

    if _cuda_lib is None:
        return
    try:
        _cu_check_error(_cuda_lib.cuInit(0))
    except NVDeviceAPIError as ex:
        if ex.errno == CU_NO_CUDA_CAPABLE_DEVICE_DETECTED:
            _cuda_lib = None
            if not _no_device_warned:
                logger.warning("No CUDA device detected")
                _no_device_warned = True
        else:
            logger.exception("Failed to initialize libcuda.")
        return


def _init_nvml():
    global _nvml_lib, _no_device_warned
    if _init_pid == os.getpid():
        return

    nvml_paths = [
        "libnvidia-ml.so",
        "libnvidia-ml.so.1",
        "libnvidia-ml.dylib",
        "nvml.dll",
    ]
    if _is_windows:
        nvml_paths.append(
            os.path.join(
                os.getenv("ProgramFiles", "C:/Program Files"),
                "NVIDIA Corporation/NVSMI/nvml.dll",
            )
        )
    if _is_wsl:
        nvml_paths = ["/usr/lib/wsl/lib/libnvidia-ml.so.1"] + nvml_paths
    _nvml_lib = _load_nv_library(*nvml_paths)

    if _nvml_lib is None:
        return
    try:
        _nvml_check_error(_nvml_lib.nvmlInit_v2())
    except NVMLAPIError as ex:
        if ex.errno == NVML_DRIVER_NOT_LOADED:
            _nvml_lib = None
            if not _no_device_warned:
                logger.warning(
                    "Failed to load libnvidia-ml: %s, no CUDA device will be enabled",
                    ex.message,
                )
                _no_device_warned = True
        else:
            logger.exception("Failed to initialize libnvidia-ml.")
        return


def _init():
    global _init_pid

    _init_cp()
    _init_nvml()

    if _nvml_lib is not None and _cuda_lib is not None:
        _init_pid = os.getpid()


def get_device_count():
    global _gpu_count

    if _gpu_count is not None:
        return _gpu_count

    _init_nvml()
    if _nvml_lib is None:
        return None

    if "CUDA_VISIBLE_DEVICES" in os.environ:
        devices = os.environ["CUDA_VISIBLE_DEVICES"].strip()
        if not devices:
            _gpu_count = 0
        else:
            _gpu_count = len(devices.split(","))
    else:
        n_gpus = c_uint()
        _cu_check_error(_nvml_lib.nvmlDeviceGetCount(byref(n_gpus)))
        _gpu_count = n_gpus.value
    return _gpu_count


def get_driver_info():
    global _driver_info

    _init_nvml()
    if _nvml_lib is None:
        return None
    if _driver_info is not None:
        return _driver_info

    version_buf = create_string_buffer(100)
    cuda_version = c_uint()

    _nvml_check_error(
        _nvml_lib.nvmlSystemGetDriverVersion(version_buf, len(version_buf))
    )
    _nvml_check_error(_nvml_lib.nvmlSystemGetCudaDriverVersion(byref(cuda_version)))

    _driver_info = _nvml_driver_info(
        driver_version=version_buf.value.decode(),
        cuda_version=".".join(str(v) for v in divmod(cuda_version.value, 1000)),
    )
    return _driver_info


def get_device_info(dev_index):
    try:
        return _device_infos[dev_index]
    except KeyError:
        pass

    _init()
    if _init_pid is None:
        return None

    device = c_int()
    name_buf = create_string_buffer(100)
    uuid_t = _CUuuid_t()
    cc_major = c_int()
    cc_minor = c_int()
    cores = c_int()
    threads_per_core = c_int()

    _cu_check_error(_cuda_lib.cuDeviceGet(byref(device), c_int(dev_index)))
    _cu_check_error(_cuda_lib.cuDeviceGetName(name_buf, len(name_buf), device))
    _cu_check_error(_cuda_lib.cuDeviceGetUuid(byref(uuid_t), device))
    _cu_check_error(
        _cuda_lib.cuDeviceComputeCapability(byref(cc_major), byref(cc_minor), device)
    )
    _cu_check_error(
        _cuda_lib.cuDeviceGetAttribute(
            byref(cores), CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT, device
        )
    )
    _cu_check_error(
        _cuda_lib.cuDeviceGetAttribute(
            byref(threads_per_core),
            CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_MULTIPROCESSOR,
            device,
        )
    )

    if "CUDA_VISIBLE_DEVICES" in os.environ:
        real_dev_index = [
            int(s) for s in os.environ["CUDA_VISIBLE_DEVICES"].split(",")
        ][dev_index]
    else:
        real_dev_index = dev_index

    info = _device_infos[dev_index] = _cu_device_info(
        index=real_dev_index,
        uuid=uuid.UUID(bytes=uuid_t.bytes),
        name=name_buf.value.decode(),
        multiprocessors=cores.value,
        cuda_cores=cores.value
        * _cu_get_processor_cores(cc_major.value, cc_minor.value),
        threads=cores.value * threads_per_core.value,
    )
    return info


def get_device_status(dev_index):
    _init()
    if _init_pid is None:
        return None

    c_device = _nvmlDevice_t()
    c_utils = _nvmlUtilization_t()
    c_temperature = c_uint()
    c_memory_info = _nvmlBAR1Memory_t()

    dev_uuid = get_device_info(dev_index).uuid

    uuid_str = ("GPU-" + str(dev_uuid)).encode()

    if not _is_wsl:
        _nvml_check_error(
            _nvml_lib.nvmlDeviceGetHandleByUUID(uuid_str, byref(c_device))
        )

        _nvml_check_error(
            _nvml_lib.nvmlDeviceGetUtilizationRates(c_device, byref(c_utils))
        )
        gpu_util = c_utils.gpu
        mem_util = c_utils.memory

        _nvml_check_error(
            _nvml_lib.nvmlDeviceGetTemperature(
                c_device, NVML_TEMPERATURE_GPU, byref(c_temperature)
            )
        )
        temperature = c_temperature.value

        _nvml_check_error(
            _nvml_lib.nvmlDeviceGetMemoryInfo(c_device, byref(c_memory_info))
        )
        fb_total_mem = c_memory_info.total
        fb_free_mem = c_memory_info.free
        fb_used_mem = c_memory_info.used
    else:
        import defusedxml

        proc = subprocess.Popen(
            ["/usr/lib/wsl/lib/nvidia-smi", "-q", f"--id={dev_index}", "-x"],
            stdout=subprocess.PIPE,
        )
        proc.wait()
        xml_result = defusedxml.ElementTree.fromstring(proc.stdout.read())
        gpu_node = xml_result.find("gpu")

        fb_node = gpu_node.find("fb_memory_usage")
        fb_total_mem = int(parse_readable_size(fb_node.find("total").text)[0])
        fb_free_mem = int(parse_readable_size(fb_node.find("free").text)[0])
        fb_used_mem = int(parse_readable_size(fb_node.find("used").text)[0])

        util_node = gpu_node.find("utilization")
        if util_node.find("gpu_util").text == "N/A":
            gpu_util = 0
        else:
            gpu_util = int(util_node.find("gpu_util"))
        if util_node.find("memory_util").text == "N/A":
            mem_util = 0
        else:
            mem_util = int(util_node.find("memory_util"))

        temperature = int(gpu_node.find("temperature").find("gpu_temp").text[:-1])

    return _nvml_device_status(
        gpu_util=gpu_util,
        mem_util=mem_util,
        temperature=temperature,
        fb_total_mem=fb_total_mem,
        fb_free_mem=fb_free_mem,
        fb_used_mem=fb_used_mem,
    )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import AsyncGenerator, List, Optional, Union

from ..core import NodeRole
from .base import AbstractClusterBackend, register_cluster_backend


@register_cluster_backend
class FixedClusterBackend(AbstractClusterBackend):
    name = "fixed"

    def __init__(self, lookup_address: Union[List[str], str]):
        if isinstance(lookup_address, str):
            lookup_address = lookup_address.split(",")
        self._supervisors = [n.strip() for n in lookup_address]

    @classmethod
    async def create(
        cls, node_role: NodeRole, lookup_address: Optional[str], pool_address: str
    ):
        return cls(lookup_address)

    async def watch_supervisors(self) -> AsyncGenerator[List[str], None]:
        yield self._supervisors

    async def get_supervisors(self, filter_ready: bool = True) -> List[str]:
        return self._supervisors

    async def request_worker(
        self, worker_cpu: int = None, worker_mem: int = None, timeout: int = None
    ) -> str:
        raise NotImplementedError

    async def release_worker(self, address: str):
        raise NotImplementedError

    async def reconstruct_worker(self, address: str):
        raise NotImplementedError



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .dmatrix import MarsDMatrix
from .train import train
from .predict import predict


def register_op():
    from .start_tracker import StartTracker

    del StartTracker


from ..utils import config_mod_getattr as _config_mod_getattr

_config_mod_getattr(
    {
        "XGBClassifier": ".classifier.XGBClassifier",
        "XGBRegressor": ".regressor.XGBRegressor",
    },
    globals(),
)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from enum import Enum

from .fuse import FUSE_CHUNK_TYPE
from .objects import OBJECT_TYPE, OBJECT_CHUNK_TYPE


class OutputType(Enum):
    object = 1
    tensor = 2
    dataframe = 3
    series = 4
    index = 5
    scalar = 6
    categorical = 7
    dataframe_groupby = 8
    series_groupby = 9

    @classmethod
    def serialize_list(cls, output_types):
        return [ot.value for ot in output_types] if output_types is not None else None

    @classmethod
    def deserialize_list(cls, output_types):
        return [cls(ot) for ot in output_types] if output_types is not None else None


_OUTPUT_TYPE_TO_CHUNK_TYPES = {OutputType.object: OBJECT_CHUNK_TYPE}
_OUTPUT_TYPE_TO_TILEABLE_TYPES = {OutputType.object: OBJECT_TYPE}
_OUTPUT_TYPE_TO_FETCH_CLS = {}


def register_output_types(output_type, tileable_types, chunk_types):
    _OUTPUT_TYPE_TO_TILEABLE_TYPES[output_type] = tileable_types
    _OUTPUT_TYPE_TO_CHUNK_TYPES[output_type] = chunk_types


def register_fetch_class(output_type, fetch_cls, fetch_shuffle_cls):
    _OUTPUT_TYPE_TO_FETCH_CLS[output_type] = (fetch_cls, fetch_shuffle_cls)


def get_tileable_types(output_type):
    return _OUTPUT_TYPE_TO_TILEABLE_TYPES[output_type]


def get_chunk_types(output_type):
    return _OUTPUT_TYPE_TO_CHUNK_TYPES[output_type]


def get_fetch_class(output_type):
    return _OUTPUT_TYPE_TO_FETCH_CLS[output_type]


@functools.lru_cache(100)
def _get_output_type_by_cls(cls):
    for tp in OutputType.__members__.values():
        try:
            tileable_types = _OUTPUT_TYPE_TO_TILEABLE_TYPES[tp]
            chunk_types = _OUTPUT_TYPE_TO_CHUNK_TYPES[tp]
            if issubclass(cls, (tileable_types, chunk_types)):
                return tp
        except KeyError:  # pragma: no cover
            continue
    raise TypeError("Output can only be tensor, dataframe or series")


def get_output_types(*objs, unknown_as=None):
    output_types = []
    for obj in objs:
        if obj is None:
            continue
        elif isinstance(obj, FUSE_CHUNK_TYPE):
            obj = obj.chunk

        try:
            output_types.append(_get_output_type_by_cls(type(obj)))
        except TypeError:
            if unknown_as is not None:
                output_types.append(unknown_as)
            else:  # pragma: no cover
                raise
    return output_types



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import itertools
import warnings
from collections import defaultdict
from typing import Callable, Iterable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from sklearn.base import (
    BaseEstimator,
    ClassifierMixin,
    RegressorMixin,
    clone as clone_estimator,
)
from sklearn.utils import check_random_state as sklearn_check_random_state

from ..utils import column_or_1d, convert_to_tensor_or_dataframe
from ..utils.multiclass import check_classification_targets
from ..utils.validation import check_is_fitted
from ... import opcodes, tensor as mt
from ...core import OutputType, get_output_types, recursive_tile
from ...core.context import Context
from ...core.operand import OperandStage
from ...dataframe.core import DATAFRAME_TYPE
from ...dataframe.utils import parse_index
from ...deploy.oscar.session import execute
from ...serialization.serializables import (
    AnyField,
    BoolField,
    DictField,
    Int8Field,
    Int64Field,
    Float32Field,
    TupleField,
    FunctionField,
    ReferenceField,
    FieldTypes,
)
from ...tensor.core import TENSOR_CHUNK_TYPE
from ...tensor.random import RandomStateField
from ...tensor.utils import gen_random_seeds
from ...typing import TileableType
from ...utils import has_unknown_shape
from ..operands import LearnOperand, LearnOperandMixin, LearnShuffleProxy
from ..utils.shuffle import LearnShuffle


def _extract_bagging_io(io_list: Iterable, op: LearnOperand, output: bool = False):
    if not isinstance(io_list, Iterable):
        io_list = [io_list]
    input_iter = iter(io_list)
    out = [
        next(input_iter),
        next(input_iter) if op.with_labels else None,
        next(input_iter) if op.with_weights else None,
        next(input_iter) if output and op.with_feature_indices else None,
    ]
    return out


def _get_by_iloc(x, idx, axis=0):
    if hasattr(x, "iloc"):
        item_getter = x.iloc
    else:
        item_getter = x
    if axis == 0:
        return item_getter[idx]
    else:
        return item_getter[:, idx]


def _concat_on_axis(data_list, axis=0, out_chunk=None):
    if isinstance(out_chunk, TENSOR_CHUNK_TYPE):
        return np.concatenate(data_list, axis=axis)
    else:
        return pd.concat(data_list, axis=axis)


def _concat_by_row(row, out_chunk=None):
    arr = np.empty((1,), dtype=object)
    arr[0] = _concat_on_axis(row.tolist(), axis=0, out_chunk=out_chunk)
    return arr


def _set_random_states(estimator, random_state=None):
    random_state = sklearn_check_random_state(random_state)
    to_set = {}
    for key in sorted(estimator.get_params(deep=True)):
        if key == "random_state" or key.endswith("__random_state"):
            to_set[key] = random_state.randint(np.iinfo(np.int32).max)

    if to_set:
        estimator.set_params(**to_set)


def _make_estimator(estimator, random_state=None):
    """Make and configure a copy of the `base_estimator_` attribute.

    Warning: This method should be used to properly instantiate new
    sub-estimators.
    """
    estimator = clone_estimator(estimator)
    if random_state is not None:
        _set_random_states(estimator, random_state)
    return estimator


class BaggingSample(LearnShuffle, LearnOperandMixin):
    _op_type_ = opcodes.BAGGING_SHUFFLE_SAMPLE

    n_estimators: int = Int64Field("n_estimators")
    max_samples = AnyField("max_samples")
    max_features = AnyField("max_features")
    bootstrap: bool = BoolField("bootstrap")
    bootstrap_features: bool = BoolField("bootstrap_features")

    random_state = RandomStateField("random_state")
    sample_random_state = RandomStateField("sample_random_state")
    feature_random_state = RandomStateField("feature_random_state")

    reducer_ratio: float = Float32Field("reducer_ratio")
    column_offset: int = Int64Field("column_offset", default=None)

    chunk_shape: Tuple[int] = TupleField("chunk_shape", FieldTypes.int64)
    with_labels: bool = BoolField("with_labels")
    with_weights: bool = BoolField("with_weights")
    with_feature_indices: bool = BoolField("with_feature_indices")

    def __init__(
        self,
        max_samples: Union[int, float] = 1.0,
        max_features: Union[int, float] = 1.0,
        bootstrap: bool = True,
        bootstrap_features: bool = False,
        random_state: np.random.RandomState = None,
        reducer_ratio: float = 1.0,
        **kw,
    ):
        super().__init__(
            bootstrap=bootstrap,
            bootstrap_features=bootstrap_features,
            max_samples=max_samples,
            max_features=max_features,
            reducer_ratio=reducer_ratio,
            random_state=random_state,
            **kw,
        )
        if self.random_state is None:
            self.random_state = np.random.RandomState()

    @property
    def output_limit(self) -> int:
        if self.stage != OperandStage.map:
            return 1 + self.with_labels + self.with_weights + self.with_feature_indices
        return 1

    def __call__(
        self,
        in_sample: TileableType,
        in_labels: Optional[TileableType] = None,
        in_weights: Optional[TileableType] = None,
    ):
        self._output_types = get_output_types(in_sample, in_labels, in_weights)

        self.with_labels = in_labels is not None
        self.with_weights = in_weights is not None
        axis_keep_shape = [
            isinstance(self.max_samples, float) and self.max_samples == 1.0,
            isinstance(self.max_features, float) and self.max_features == 1.0,
        ]
        self.with_feature_indices = not axis_keep_shape[1] or self.bootstrap_features
        if self.with_feature_indices:
            self._output_types += (OutputType.tensor,)

        new_shape = tuple(
            s if keep_shape else np.nan
            for s, keep_shape in zip(in_sample.shape, axis_keep_shape)
        )

        kws = []

        data_params = in_sample.params
        data_params["shape"] = new_shape
        kws.append(data_params)

        if in_labels is not None:
            labels_params = in_labels.params
            labels_params["shape"] = (new_shape[0],)
            kws.append(labels_params)

        if in_weights is not None:
            weights_params = in_weights.params
            weights_params["shape"] = (new_shape[0],)
            kws.append(weights_params)

        if self.with_feature_indices:
            feature_params = {
                "shape": (self.n_estimators, new_shape[1]),
                "dtype": np.dtype(int),
            }
            kws.append(feature_params)

        inputs = [in_sample]
        if in_labels is not None:
            inputs.append(in_labels)
        if in_weights is not None:
            inputs.append(in_weights)

        return self.new_tileables(inputs, kws=kws)

    @classmethod
    def _scatter_samples(
        cls,
        max_samples: Union[int, float],
        nsplits: Tuple[int],
        random_state: np.random.RandomState,
        n_estimators: int,
    ) -> np.ndarray:
        nsp_array = np.array(nsplits)
        dim_size = nsp_array.sum()
        if isinstance(max_samples, int):
            expect_sample_count = max_samples
        else:
            expect_sample_count = int(max_samples * nsp_array.sum())

        if expect_sample_count == dim_size:
            return np.array([list(nsplits)] * n_estimators)

        split_probs = nsp_array / dim_size
        return random_state.multinomial(
            expect_sample_count, split_probs, size=n_estimators
        )

    @classmethod
    def tile(cls, op: "BaggingSample"):
        in_sample, in_labels, in_weights, _ = _extract_bagging_io(
            op.inputs, op, output=False
        )
        out_data, out_labels, out_weights, out_feature_indices = _extract_bagging_io(
            op.outputs, op, output=True
        )

        # make sure all shapes are computed
        if (
            has_unknown_shape(in_sample)
            or (in_labels is not None and has_unknown_shape(in_labels))
            or (in_weights is not None and has_unknown_shape(in_weights))
        ):
            yield

        to_tile = []
        if in_labels is not None:
            in_labels = in_labels.rechunk({0: in_sample.nsplits[0]})
            to_tile.append(in_labels)
        if in_weights is not None:
            in_weights = in_weights.rechunk({0: in_sample.nsplits[0]})
            to_tile.append(in_weights)

        # tile rechunks
        if to_tile:
            tiled = yield from recursive_tile(to_tile)
            tiled_iter = iter(tiled)
            if in_labels is not None:
                in_labels = next(tiled_iter)
            if in_weights is not None:
                in_weights = next(tiled_iter)

        random_seeds = [
            gen_random_seeds(n, op.random_state) for n in in_sample.chunk_shape
        ]

        axis_keep_shape = [
            isinstance(op.max_samples, float)
            and op.max_samples == 1.0
            and not op.bootstrap,
            isinstance(op.max_features, float)
            and op.max_features == 1.0
            and not op.bootstrap_features,
        ]

        n_reducers = (
            op.n_reducers
            if getattr(op, "n_reducers", None)
            else max(1, int(in_sample.chunk_shape[0] * op.reducer_ratio))
        )

        # todo implement sampling without replacements
        map_chunks = []
        max_samples_splits = cls._scatter_samples(
            op.max_samples, in_sample.nsplits[0], op.random_state, op.n_estimators
        )
        max_features_splits = cls._scatter_samples(
            op.max_features, in_sample.nsplits[1], op.random_state, op.n_estimators
        )

        column_cum_offset = np.concatenate([[0], np.cumsum(in_sample.nsplits[1])])
        for chunk in in_sample.chunks:
            new_op = op.copy().reset_key()
            new_op.random_state = None
            new_op.sample_random_state = np.random.RandomState(
                random_seeds[0][chunk.index[0]]
            )
            new_op.feature_random_state = np.random.RandomState(
                random_seeds[1][chunk.index[1]]
            )
            new_op.stage = OperandStage.map
            new_op.max_samples = max_samples_splits[:, chunk.index[0]]
            new_op.max_features = max_features_splits[:, chunk.index[1]]
            new_op.n_reducers = n_reducers
            new_op.column_offset = int(column_cum_offset[chunk.index[1]])

            if chunk.index[0] != 0:
                new_op.with_feature_indices = False

            if chunk.index[1] != in_sample.chunk_shape[1] - 1:
                new_op.with_weights = False
                new_op.with_labels = False

            params = chunk.params
            params["shape"] = tuple(
                s if keep_shape else np.nan
                for s, keep_shape in zip(chunk.shape, axis_keep_shape)
            )

            input_chunks = [chunk]
            if new_op.with_labels:
                input_chunks.append(in_labels.cix[chunk.index[0]])
            if new_op.with_weights:
                input_chunks.append(in_weights.cix[chunk.index[0]])
            map_chunks.append(new_op.new_chunk(input_chunks, **params))

        shuffle_op = LearnShuffleProxy(output_types=[OutputType.tensor]).new_chunk(
            map_chunks, dtype=np.dtype(int), shape=()
        )

        remain_reducers = op.n_estimators % n_reducers
        reduce_data_chunks = []
        reduce_labels_chunks = []
        reduce_weights_chunks = []
        reduce_feature_chunks = []
        for idx in range(n_reducers):
            new_op = op.copy().reset_key()
            new_op.random_state = None
            new_op.stage = OperandStage.reduce
            new_op.reducer_ordinal = idx
            new_op.n_reducers = n_reducers
            new_op.chunk_shape = in_sample.chunk_shape
            new_op.n_estimators = op.n_estimators // n_reducers
            if remain_reducers:
                remain_reducers -= 1
                new_op.n_estimators += 1

            if new_op.n_estimators == 0:
                continue

            kws = []

            data_params = out_data.params
            data_params["index"] = (idx, 0)
            data_params["shape"] = (np.nan, out_data.shape[1])
            kws.append(data_params)

            if op.with_labels:
                labels_params = out_labels.params
                labels_params["index"] = (idx,)
                labels_params["shape"] = (np.nan,)
                kws.append(labels_params)

            if op.with_weights:
                weights_params = out_weights.params
                weights_params["index"] = (idx,)
                weights_params["shape"] = (np.nan,)
                kws.append(weights_params)

            if op.with_feature_indices:
                feature_params = {
                    "index": (idx, 0),
                    "shape": (new_op.n_estimators, out_feature_indices.shape[1]),
                    "dtype": np.dtype(int),
                }
                kws.append(feature_params)

            chunks = new_op.new_chunks([shuffle_op], kws=kws)
            (
                data_chunk,
                labels_chunk,
                weights_chunk,
                feature_chunk,
            ) = _extract_bagging_io(chunks, op, output=True)

            reduce_data_chunks.append(data_chunk)
            if labels_chunk is not None:
                reduce_labels_chunks.append(labels_chunk)
            if weights_chunk is not None:
                reduce_weights_chunks.append(weights_chunk)
            if feature_chunk is not None:
                reduce_feature_chunks.append(feature_chunk)

        new_op = op.copy().reset_key()

        kws = [
            {
                "chunks": reduce_data_chunks,
                "nsplits": ((np.nan,) * len(reduce_data_chunks), (out_data.shape[1],)),
                **out_data.params,
            }
        ]
        if op.with_labels:
            kws.append(
                {
                    "chunks": reduce_labels_chunks,
                    "nsplits": ((np.nan,) * len(reduce_data_chunks),),
                    **out_labels.params,
                }
            )
        if op.with_weights:
            kws.append(
                {
                    "chunks": reduce_weights_chunks,
                    "nsplits": ((np.nan,) * len(reduce_data_chunks),),
                    **out_weights.params,
                }
            )
        if op.with_feature_indices:
            estimator_nsplit = tuple(c.op.n_estimators for c in reduce_data_chunks)
            kws.append(
                {
                    "chunks": reduce_feature_chunks,
                    "nsplits": (estimator_nsplit, (out_feature_indices.shape[1],)),
                    **out_feature_indices.params,
                }
            )
        return new_op.new_tileables(op.inputs, kws=kws)

    @classmethod
    def _gen_sample_indices(
        cls,
        max_range: int,
        size: int,
        random_state: np.random.RandomState,
        with_replacement: bool = False,
    ):
        if not with_replacement:
            result = random_state.choice(np.arange(max_range), size, False)
        else:
            result = random_state.randint(0, max_range - 1, size)
        result.sort()
        return result

    @classmethod
    def _execute_map(cls, ctx, op: "BaggingSample"):
        in_sample, in_labels, in_weights, _ = _extract_bagging_io(
            op.inputs, op, output=False
        )
        in_sample_data = ctx[in_sample.key]
        in_labels_data = ctx[in_labels.key] if op.with_labels else None
        in_weights_data = ctx[in_weights.key] if op.with_weights else None
        out_samples = op.outputs[0]

        remains = op.n_estimators % op.n_reducers
        reducer_iters = [
            itertools.repeat(idx, 1 + op.n_estimators // op.n_reducers)
            for idx in range(remains)
        ]
        reducer_iters += [
            itertools.repeat(idx, op.n_estimators // op.n_reducers)
            for idx in range(remains, op.n_reducers)
        ]
        reducer_iter = itertools.chain(*reducer_iters)

        result_store = defaultdict(lambda: ([], [], [], []))
        for est_id in range(op.n_estimators):
            sampled_data = in_sample_data
            sampled_labels = in_labels_data
            sampled_weights = in_weights_data

            if op.max_samples[est_id] != in_sample_data.shape[0]:
                sample_indices = cls._gen_sample_indices(
                    in_sample_data.shape[0],
                    op.max_samples[est_id],
                    op.sample_random_state,
                    op.bootstrap,
                )

                sampled_data = _get_by_iloc(sampled_data, sample_indices)
                if sampled_labels is not None:
                    sampled_labels = _get_by_iloc(sampled_labels, sample_indices)
                if sampled_weights is not None:
                    sampled_weights = _get_by_iloc(sampled_weights, sample_indices)

            if op.max_features[est_id] != in_sample_data.shape[1]:
                feature_indices = cls._gen_sample_indices(
                    in_sample_data.shape[1],
                    op.max_features[est_id],
                    op.feature_random_state,
                    op.bootstrap_features,
                )

                sampled_data = _get_by_iloc(sampled_data, feature_indices, axis=1)
                if not op.with_feature_indices:
                    feature_indices = None
            else:
                feature_indices = None

            samples, labels, weights, feature_idx_array = result_store[
                next(reducer_iter)
            ]
            samples.append(sampled_data)
            if sampled_labels is not None:
                labels.append(sampled_labels)
            if sampled_weights is not None:
                weights.append(sampled_weights)
            if feature_indices is not None:
                feature_idx_array.append(feature_indices + op.column_offset)

        for (
            reducer_id,
            (
                samples,
                labels,
                weights,
                feature_idx_array,
            ),
        ) in result_store.items():
            ctx[out_samples.key, (reducer_id, 0)] = tuple(
                samples + labels + weights + feature_idx_array
            )

    @classmethod
    def _execute_reduce(cls, ctx, op: "BaggingSample"):
        out_data, out_labels, out_weights, out_feature_indices = _extract_bagging_io(
            op.outputs, op, output=True
        )

        input_keys = op.inputs[0].op.source_keys
        input_idxes = op.inputs[0].op.source_idxes

        sample_holder = [
            np.empty(op.chunk_shape, dtype=object) for _ in range(op.n_estimators)
        ]

        labels_holder = (
            [np.empty(op.chunk_shape[0], dtype=object) for _ in range(op.n_estimators)]
            if op.with_labels
            else None
        )

        weights_holder = (
            [np.empty(op.chunk_shape[0], dtype=object) for _ in range(op.n_estimators)]
            if op.with_weights
            else None
        )

        feature_indices_holder = (
            [np.empty(op.chunk_shape[1], dtype=object) for _ in range(op.n_estimators)]
            if op.with_feature_indices
            else None
        )

        for input_key, input_idx in zip(input_keys, input_idxes):
            add_feature_index = input_idx[0] == 0
            add_label_weight = input_idx[1] == op.chunk_shape[1] - 1
            chunk_data = ctx[input_key, out_data.index]

            num_groups = 1
            if add_feature_index and op.with_feature_indices:
                # contains feature indices
                num_groups += 1
            if add_label_weight:  # contains label or weight
                num_groups += int(op.with_weights) + int(op.with_labels)

            sample_count = len(chunk_data) // num_groups
            assert len(chunk_data) % num_groups == 0

            group_iter = (
                chunk_data[i * sample_count : (i + 1) * sample_count]
                for i in range(num_groups)
            )

            for data_idx, sample in enumerate(next(group_iter)):
                sample_holder[data_idx][input_idx] = sample

            if add_label_weight:
                if op.with_labels:
                    for data_idx, label in enumerate(next(group_iter)):
                        labels_holder[data_idx][input_idx[0]] = label
                if op.with_weights:
                    for data_idx, weight in enumerate(next(group_iter)):
                        weights_holder[data_idx][input_idx[0]] = weight

            if add_feature_index and op.with_feature_indices:
                for data_idx, feature_index in enumerate(next(group_iter)):
                    feature_indices_holder[data_idx][input_idx[1]] = feature_index

        data_results: List[Optional[np.ndarray]] = [None] * len(sample_holder)
        for est_idx, sample_mat in enumerate(sample_holder):
            row_chunks = np.apply_along_axis(
                _concat_by_row, axis=0, arr=sample_mat, out_chunk=out_data
            )
            data_results[est_idx] = _concat_on_axis(
                row_chunks[0].tolist(), axis=1, out_chunk=out_data
            )
        ctx[out_data.key] = tuple(data_results)

        for out, holder in zip(
            (out_labels, out_weights, out_feature_indices),
            (labels_holder, weights_holder, feature_indices_holder),
        ):
            if out is None:
                continue
            results: List[Optional[np.ndarray]] = [None] * len(holder)
            for est_idx, labels_vct in enumerate(holder):
                results[est_idx] = _concat_on_axis(labels_vct.tolist(), out_chunk=out)
            if holder is feature_indices_holder:
                ctx[out.key] = np.stack(results)
            else:
                ctx[out.key] = tuple(results)

    @classmethod
    def execute(cls, ctx, op: "BaggingSample"):
        if op.stage == OperandStage.map:
            cls._execute_map(ctx, op)
        else:
            cls._execute_reduce(ctx, op)


class BaggingSampleReindex(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.BAGGING_SHUFFLE_REINDEX

    n_estimators: int = Int64Field("n_estimators")
    feature_indices: TileableType = ReferenceField("feature_indices", default=None)

    start_col_index: int = Int64Field("start_col_index", 0)

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if self.feature_indices is not None:
            self.feature_indices = inputs[-1]

    def __call__(self, data: TileableType, feature_indices: TileableType = None):
        self._output_types = get_output_types(data)
        inputs = [data]
        self.feature_indices = feature_indices
        params = data.params
        if feature_indices is not None:
            inputs.append(feature_indices)
            params["shape"] = (data.shape[0], np.nan)
        if isinstance(data, DATAFRAME_TYPE):
            params["index_value"] = parse_index(pd.Index([], dtype=np.int64), data.key)
        return self.new_tileable(inputs, **params)

    @classmethod
    def tile(cls, op: "BaggingSampleReindex"):
        t_data = op.inputs[0]
        t_out = op.outputs[0]
        t_feature_idxes = op.feature_indices
        cum_nsplits = np.cumsum(np.concatenate([[0], t_data.nsplits[1]]))

        if t_feature_idxes is None:
            out = t_data
            if out.chunk_shape[1] > 1:
                out = yield from recursive_tile(out.rechunk({1: (out.shape[1],)}))
            return out

        # generate map chunks
        map_holder = np.empty(
            t_data.chunk_shape + (t_feature_idxes.chunk_shape[0],),
            dtype=np.dtype(object),
        )
        for chunk in t_data.chunks:
            for feature_idx_chunk in t_feature_idxes.chunks:
                new_op = op.copy().reset_key()
                new_op.stage = OperandStage.map
                new_op.start_col_index = int(cum_nsplits[chunk.index[1]])
                params = chunk.params
                new_index = params["index"] = chunk.index + (
                    feature_idx_chunk.index[0],
                )
                if t_feature_idxes.chunk_shape[0] == 1:
                    new_index = new_index[:-1]
                map_holder[new_index] = new_op.new_chunk(
                    [chunk, feature_idx_chunk], **params
                )
        if op.feature_indices.chunk_shape[0] == 1:
            chunks = map_holder.reshape((t_data.chunk_shape[0],)).tolist()
        else:

            def _gen_combine_chunk(chunks):
                new_op = op.copy().reset_key()
                new_op.feature_indices = None
                new_op.stage = OperandStage.combine
                params = chunks[0].params
                params["shape"] = (chunks[0].shape[0], op.feature_indices.shape[1])
                params["index"] = (chunks[0].index[0], chunks[0].index[2])
                if isinstance(t_data, DATAFRAME_TYPE):
                    params["index_value"] = parse_index(
                        pd.Index([], dtype=np.int64), chunks[0].key
                    )
                inputs = chunks.tolist()
                return new_op.new_chunk(inputs, **params)

            chunks_array = np.apply_along_axis(_gen_combine_chunk, 1, map_holder)
            chunks = chunks_array.reshape((chunks_array.size,)).tolist()

        new_op = op.copy().reset_key()
        new_nsplits = (
            t_data.nsplits[0],
            (op.feature_indices.shape[1],) * t_feature_idxes.chunk_shape[0],
        )
        return new_op.new_tileables(
            op.inputs, chunks=chunks, nsplits=new_nsplits, **t_out.params
        )

    @classmethod
    def _execute_map(cls, ctx, op: "BaggingSampleReindex"):
        data = ctx[op.inputs[0].key]
        feature_idx = ctx[op.feature_indices.key] - op.start_col_index
        filtered = []
        for row in feature_idx:
            row = row[(row >= 0) & (row < data.shape[1])]
            filtered.append(_get_by_iloc(data, row, axis=1))
        ctx[op.outputs[0].key] = tuple(filtered)

    @classmethod
    def _execute_combine(cls, ctx, op: "BaggingSampleReindex"):
        data_inputs = [ctx[c.key] for c in op.inputs]
        concatenated = []
        for data_input in zip(*data_inputs):
            concatenated.append(_concat_on_axis(data_input, 1, op.inputs[0]))
        ctx[op.outputs[0].key] = tuple(concatenated)

    @classmethod
    def execute(cls, ctx, op: "BaggingSampleReindex"):
        if op.stage == OperandStage.combine:
            cls._execute_combine(ctx, op)
        else:
            cls._execute_map(ctx, op)


class BaggingFitOperand(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.BAGGING_FIT

    base_estimator: BaseEstimator = AnyField("base_estimator")
    estimator_params: dict = DictField("estimator_params", default=None)
    n_estimators: int = Int64Field("n_estimators")
    max_samples = AnyField("max_samples", default=1.0)
    max_features = AnyField("max_features", default=1.0)
    bootstrap: bool = BoolField("bootstrap", default=False)
    bootstrap_features: bool = BoolField("bootstrap_features", default=True)
    random_state = RandomStateField("random_state", default=None)

    reducer_ratio: float = Float32Field("reducer_ratio")
    n_reducers: int = Int64Field("n_reducers")

    labels: TileableType = ReferenceField("labels", default=None)
    weights: TileableType = ReferenceField("weights", default=None)
    feature_indices: TileableType = ReferenceField("feature_indices", default=None)
    with_feature_indices: bool = BoolField("with_feature_indices", default=None)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.random_state is None:
            self.random_state = np.random.RandomState()
        if self.with_feature_indices is None:
            full_features = (
                isinstance(self.max_features, float) and self.max_features == 1.0
            )
            self.with_feature_indices = not full_features or self.bootstrap_features

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)

        input_iter = iter(inputs)
        next(input_iter)
        if self.labels is not None:
            self.labels = next(input_iter)
        if self.weights is not None:
            self.weights = next(input_iter)
        if self.feature_indices is not None:
            self.feature_indices = next(input_iter)

    def _get_bagging_sample_tileables(self, samples=None):
        samples = samples or self.inputs[0]
        sample_op = BaggingSample(
            n_estimators=self.n_estimators,
            max_samples=self.max_samples,
            max_features=self.max_features,
            bootstrap=self.bootstrap,
            bootstrap_features=self.bootstrap_features,
            random_state=self.random_state,
            reducer_ratio=self.reducer_ratio,
            n_reducers=self.n_reducers,
            with_weights=self.weights is not None,
            with_labels=self.labels is not None,
            with_feature_indices=self.with_feature_indices,
        )
        return _extract_bagging_io(
            sample_op(samples, self.labels, self.weights), sample_op, output=True
        )

    @property
    def output_limit(self) -> int:
        if self.with_feature_indices:
            return 2
        return 1

    def __call__(
        self,
        in_data: TileableType,
        in_labels: Optional[TileableType] = None,
        in_weights: Optional[TileableType] = None,
        feature_indices: TileableType = None,
    ):
        self._output_types = [OutputType.tensor]
        inputs = [in_data]

        if in_labels is not None:
            self.labels = in_labels
            inputs.append(in_labels)
        if in_weights is not None:
            self.weights = in_weights
            inputs.append(in_weights)

        if feature_indices is not None:
            self.feature_indices = feature_indices
            inputs.append(feature_indices)

        kws = [dict(shape=(self.n_estimators,), dtype=np.dtype(object))]
        if self.with_feature_indices:
            self._output_types.append(OutputType.tensor)
            sample_tileables = self._get_bagging_sample_tileables(in_data)
            kws.append(sample_tileables[-1].params)

        return self.new_tileables(inputs, kws=kws)

    @classmethod
    def tile(cls, op: "BaggingFitOperand"):
        out = op.outputs[0]
        sample_tileables = op._get_bagging_sample_tileables()
        tiled_sample_iter = iter(
            (
                yield from recursive_tile(
                    tuple(t for t in sample_tileables if t is not None)
                )
            )
        )
        sampled, labels, weights, feature_indices = (
            t if t is None else next(tiled_sample_iter) for t in sample_tileables
        )

        estimator_nsplits = (tuple(c.op.n_estimators for c in sampled.chunks),)

        label_chunks = itertools.repeat(None) if labels is None else labels.chunks
        weight_chunks = itertools.repeat(None) if weights is None else weights.chunks

        out_chunks = []
        seeds = gen_random_seeds(len(sampled.chunks), op.random_state)
        for sample_chunk, label_chunk, weight_chunk, n_estimators in zip(
            sampled.chunks, label_chunks, weight_chunks, estimator_nsplits[0]
        ):
            chunk_op = BaggingFitOperand(
                base_estimator=op.base_estimator,
                estimator_params=op.estimator_params,
                labels=label_chunk,
                weights=weight_chunk,
                n_estimators=n_estimators,
                with_feature_indices=False,
                random_state=sklearn_check_random_state(seeds[sample_chunk.index[0]]),
            )
            chunk_op._output_types = op._output_types
            inputs = [
                c for c in [sample_chunk, label_chunk, weight_chunk] if c is not None
            ]
            out_chunks.append(
                chunk_op.new_chunk(
                    inputs,
                    index=(sample_chunk.index[0],),
                    shape=(n_estimators,),
                    dtype=out.dtype,
                )
            )

        out_op = op.copy().reset_key()
        kws = [
            dict(chunks=out_chunks, nsplits=estimator_nsplits, **out.params),
        ]
        if feature_indices is not None:
            kws.append(
                dict(
                    chunks=feature_indices.chunks,
                    nsplits=feature_indices.nsplits,
                    **feature_indices.params,
                )
            )
        return out_op.new_tileables(op.inputs, kws=kws, output_limit=len(kws))

    @classmethod
    def execute(cls, ctx: Union[dict, Context], op: "BaggingFitOperand"):
        sampled_data = ctx[op.inputs[0].key]
        labels_data = (
            ctx[op.labels.key] if op.labels is not None else itertools.repeat(None)
        )
        weights_data = (
            ctx[op.weights.key] if op.weights is not None else itertools.repeat(None)
        )

        for k, v in (op.estimator_params or dict()).items():
            setattr(op.base_estimator, k, v)

        new_estimators = []
        seeds = gen_random_seeds(len(sampled_data), op.random_state)
        for idx, (sampled, label, weights) in enumerate(
            zip(sampled_data, labels_data, weights_data)
        ):
            estimator = _make_estimator(op.base_estimator, seeds[idx])
            estimator.fit(sampled, y=label, sample_weight=weights)
            new_estimators.append(estimator)
        ctx[op.outputs[0].key] = np.array(new_estimators, dtype=np.dtype(object))


class PredictionType(enum.Enum):
    REGRESSION = 0
    PROBABILITY = 1
    LOG_PROBABILITY = 2
    DECISION_FUNCTION = 3


class BaggingPredictionOperand(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.BAGGING_PREDICTION

    estimators: TileableType = ReferenceField("estimators")
    feature_indices: TileableType = ReferenceField("feature_indices", default=None)
    n_classes: Optional[int] = Int64Field("n_classes", default=None)
    prediction_type: PredictionType = Int8Field(
        "prediction_type",
        on_serialize=lambda x: x.value,
        on_deserialize=PredictionType,
        default=PredictionType.PROBABILITY,
    )
    decision_function: Callable = FunctionField("decision_function", default=None)
    calc_means: bool = BoolField("calc_means", default=True)

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        input_iter = iter(inputs[1:])
        self.estimators = next(input_iter)
        if self.feature_indices is not None:
            self.feature_indices = next(input_iter)

    def __call__(
        self,
        instances: TileableType,
        estimators: TileableType,
        feature_indices: TileableType = None,
    ) -> TileableType:
        self._output_types = [OutputType.tensor]
        self.estimators = estimators
        self.feature_indices = feature_indices

        if self.n_classes is not None:
            shape = (instances.shape[0], estimators.shape[0], self.n_classes)
        else:
            shape = (instances.shape[0], estimators.shape[0])
        if self.calc_means:
            shape = (shape[0],) + shape[2:]

        params = {"dtype": np.dtype(float), "shape": shape}
        inputs = [instances, estimators]
        if feature_indices is not None:
            inputs.append(feature_indices)
        return self.new_tileable(inputs, **params)

    def _get_class_shape(self):
        if self.n_classes and self.n_classes > 2:
            return self.n_classes
        elif self.prediction_type == PredictionType.DECISION_FUNCTION:
            return None
        else:
            return self.n_classes

    @classmethod
    def _build_chunks_without_feature_indices(
        cls, op: "BaggingPredictionOperand", t_instances: TileableType
    ):
        class_shape = op._get_class_shape()
        chunks = []
        for c_instance in t_instances.chunks:
            for c_estimator in op.estimators.chunks:
                if class_shape is not None:
                    params = {
                        "dtype": np.dtype(float),
                        "shape": (
                            c_instance.shape[0],
                            class_shape,
                            c_estimator.shape[0],
                        ),
                        "index": (c_instance.index[0], 0, c_estimator.index[0]),
                    }
                else:
                    params = {
                        "dtype": np.dtype(float),
                        "shape": (c_instance.shape[0], c_estimator.shape[0]),
                        "index": (c_instance.index[0], c_estimator.index[0]),
                    }
                new_op = op.copy().reset_key()
                new_op.feature_indices = None
                chunks.append(new_op.new_chunk([c_instance, c_estimator], **params))
        return chunks

    @classmethod
    def _build_chunks_with_feature_indices(
        cls, op: "BaggingPredictionOperand", t_instances: TileableType
    ):
        class_shape = op._get_class_shape()
        chunks = []
        for c in t_instances.chunks:
            estimator_chunk = op.estimators.chunks[c.index[1]]

            if class_shape is not None:
                params = {
                    "dtype": np.dtype(float),
                    "shape": (c.shape[0], class_shape, estimator_chunk.shape[0]),
                    "index": (c.index[0], 0, c.index[1]),
                }
            else:
                params = {
                    "dtype": np.dtype(float),
                    "shape": (c.shape[0], estimator_chunk.shape[0]),
                    "index": c.index,
                }

            new_op = op.copy().reset_key()
            new_op.feature_indices = None
            chunks.append(new_op.new_chunk([c, estimator_chunk], **params))
        return chunks

    @classmethod
    def tile(cls, op: "BaggingPredictionOperand"):
        n_estimators = op.estimators.shape[0]
        reindex_op = BaggingSampleReindex(n_estimators=n_estimators)
        t_instances = yield from recursive_tile(
            reindex_op(op.inputs[0], op.feature_indices)
        )

        # for classifiers, form instance-class-estimator array
        # for regressors, form instance-estimator array
        # and then sum over estimator axis

        if op.feature_indices is None:
            chunks = cls._build_chunks_without_feature_indices(op, t_instances)
        else:
            chunks = cls._build_chunks_with_feature_indices(op, t_instances)

        new_op = op.copy().reset_key()
        class_shape = op._get_class_shape()
        if class_shape is not None:
            params = {
                "dtype": np.dtype(float),
                "shape": (t_instances.shape[0], class_shape, n_estimators),
            }
            nsplits = (t_instances.nsplits[0], (class_shape,), op.estimators.nsplits[0])
        else:
            params = {
                "dtype": np.dtype(float),
                "shape": (t_instances.shape[0], n_estimators),
            }
            nsplits = (t_instances.nsplits[0], op.estimators.nsplits[0])
        estimator_probas = new_op.new_tileable(
            op.inputs, chunks=chunks, nsplits=nsplits, **params
        )

        if not op.calc_means:
            return estimator_probas
        elif op.prediction_type != PredictionType.LOG_PROBABILITY:
            return [
                (
                    yield from recursive_tile(
                        mt.sum(estimator_probas, axis=-1) / n_estimators
                    )
                )
            ]
        else:
            return [
                (
                    yield from recursive_tile(
                        mt.log(mt.exp(estimator_probas).sum(axis=-1))
                        - np.log(n_estimators)
                    )
                )
            ]

    @classmethod
    def _predict_proba(cls, instance, estimator, n_classes):
        n_samples = instance.shape[0]
        proba = np.zeros((n_samples, n_classes))

        if hasattr(estimator, "predict_proba"):
            proba_estimator = estimator.predict_proba(instance)
            if n_classes == len(estimator.classes_):
                proba += proba_estimator

            else:
                proba[:, estimator.classes_] += proba_estimator[
                    :, range(len(estimator.classes_))
                ]
        else:
            # Resort to voting
            predictions = estimator.predict(instance)
            for i in range(n_samples):
                proba[i, predictions[i]] += 1
        return proba

    @classmethod
    def _predict_log_proba(cls, instance, estimator, n_classes):
        """Private function used to compute log probabilities within a job."""
        if not hasattr(estimator, "predict_log_proba"):
            return np.log(cls._predict_proba(instance, estimator, n_classes))

        n_samples = instance.shape[0]
        log_proba = np.empty((n_samples, n_classes))
        log_proba.fill(-np.inf)
        all_classes = np.arange(n_classes, dtype=int)

        log_proba_estimator = estimator.predict_log_proba(instance)

        if n_classes == len(estimator.classes_):
            log_proba = np.logaddexp(log_proba, log_proba_estimator)
        else:  # pragma: no cover
            log_proba[:, estimator.classes_] = np.logaddexp(
                log_proba[:, estimator.classes_],
                log_proba_estimator[:, range(len(estimator.classes_))],
            )
            missing = np.setdiff1d(all_classes, estimator.classes_)
            log_proba[:, missing] = np.logaddexp(log_proba[:, missing], -np.inf)
        return log_proba

    @classmethod
    def _decision_function(cls, instance, estimator, func=None):
        if func is not None:
            return func(instance, estimator)
        else:
            return estimator.decision_function(instance)

    @classmethod
    def execute(cls, ctx: Union[dict, Context], op: "BaggingPredictionOperand"):
        instances = ctx[op.inputs[0].key]
        estimators = ctx[op.estimators.key]
        if not isinstance(instances, tuple):
            instances = [instances] * len(estimators)

        estimate_results = []
        for instance, estimator in zip(instances, estimators):
            # classifier
            if op.prediction_type == PredictionType.PROBABILITY:
                estimate_results.append(
                    cls._predict_proba(instance, estimator, op.n_classes)
                )
            elif op.prediction_type == PredictionType.LOG_PROBABILITY:
                estimate_results.append(
                    cls._predict_log_proba(instance, estimator, op.n_classes)
                )
            elif op.prediction_type == PredictionType.DECISION_FUNCTION:
                estimate_results.append(
                    cls._decision_function(instance, estimator, op.decision_function)
                )
            else:
                estimate_results.append(estimator.predict(instance))

        out = op.outputs[0]
        ctx[out.key] = np.stack(estimate_results, axis=out.ndim - 1)


class BaseBagging:
    def __init__(
        self,
        base_estimator=None,
        n_estimators=10,
        *,
        max_samples=1.0,
        max_features=1.0,
        bootstrap=True,
        bootstrap_features=False,
        oob_score=False,
        warm_start=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        reducers=1.0,
    ):
        self.base_estimator = base_estimator
        self.n_estimators = n_estimators

        self.max_samples = max_samples
        self.max_features = max_features
        self.bootstrap = bootstrap
        self.bootstrap_features = bootstrap_features
        self.oob_score = oob_score
        self.warm_start = warm_start
        self.n_jobs = n_jobs
        self.random_state = (
            np.random.RandomState(random_state)
            if isinstance(random_state, int)
            else random_state
        )
        self.verbose = verbose
        self.reducers = reducers

        self.estimators_ = None
        self.estimator_features_ = None

    def _validate_y(self, y, session=None, run_kwargs=None):
        if len(y.shape) == 1 or y.shape[1] == 1:
            return column_or_1d(y, warn=True)
        else:
            return y

    def _fit(
        self,
        X,
        y=None,
        sample_weight=None,
        max_samples=None,
        estimator_params=None,
        session=None,
        run_kwargs=None,
    ):
        estimator_features, feature_indices = None, None
        n_more_estimators = self.n_estimators

        X = convert_to_tensor_or_dataframe(X)
        y = convert_to_tensor_or_dataframe(y) if y is not None else None
        sample_weight = (
            convert_to_tensor_or_dataframe(sample_weight)
            if sample_weight is not None
            else None
        )

        y = self._validate_y(y)

        if self.warm_start:
            feature_indices = self.estimator_features_
            if self.estimators_ is not None:
                exist_estimators = self.estimators_.shape[0]
                # move random states to skip duplicated results
                self.random_state.rand(exist_estimators)
                n_more_estimators = self.n_estimators - exist_estimators

        if n_more_estimators < 0:
            raise ValueError(
                "n_estimators=%d must be larger or equal to "
                "len(estimators_)=%d when warm_start==True"
                % (self.n_estimators, self.estimators_.shape[0])
            )
        elif n_more_estimators == 0:
            warnings.warn(
                "Warm-start fitting without increasing n_estimators does not "
                "fit new trees."
            )
            return self

        fit_op = BaggingFitOperand(
            base_estimator=self.base_estimator,
            estimator_params=estimator_params,
            n_estimators=n_more_estimators,
            max_samples=max_samples or self.max_samples,
            max_features=self.max_features,
            bootstrap=self.bootstrap,
            bootstrap_features=self.bootstrap_features,
            random_state=self.random_state,
            reducer_ratio=self.reducers if isinstance(self.reducers, float) else None,
            n_reducers=self.reducers if isinstance(self.reducers, int) else None,
        )
        tileables = fit_op(X, y, sample_weight, feature_indices)
        ret = execute(*tileables, session=session, **(run_kwargs or dict()))

        if len(ret) == 2:
            estimators, estimator_features = ret
        else:
            estimators = ret

        if self.estimators_ is not None:
            estimators = mt.concatenate([self.estimators_, estimators])
        if self.estimator_features_ is not None:
            estimator_features = mt.concatenate(
                [self.estimator_features_, estimator_features]
            )

        self.estimators_, self.estimator_features_ = estimators, estimator_features
        return self

    def fit(self, X, y=None, sample_weight=None, session=None, run_kwargs=None):
        """
        Build a Bagging ensemble of estimators from the training set (X, y).

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.

        y : array-like of shape (n_samples,)
            The target values (class labels in classification, real numbers in
            regression).

        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights. If None, then samples are equally weighted.
            Note that this is supported only if the base estimator supports
            sample weighting.

        Returns
        -------
        self : object
            Fitted estimator.
        """
        return self._fit(
            X, y, sample_weight=sample_weight, session=session, run_kwargs=run_kwargs
        )


class BaggingClassifier(ClassifierMixin, BaseBagging):
    """
    A Bagging classifier.

    A Bagging classifier is an ensemble meta-estimator that fits base
    classifiers each on random subsets of the original dataset and then
    aggregate their individual predictions (either by voting or by averaging)
    to form a final prediction. Such a meta-estimator can typically be used as
    a way to reduce the variance of a black-box estimator (e.g., a decision
    tree), by introducing randomization into its construction procedure and
    then making an ensemble out of it.

    This algorithm encompasses several works from the literature. When random
    subsets of the dataset are drawn as random subsets of the samples, then
    this algorithm is known as Pasting [1]_. If samples are drawn with
    replacement, then the method is known as Bagging [2]_. When random subsets
    of the dataset are drawn as random subsets of the features, then the method
    is known as Random Subspaces [3]_. Finally, when base estimators are built
    on subsets of both samples and features, then the method is known as
    Random Patches [4]_.

    Read more in the :ref:`User Guide <bagging>`.

    Parameters
    ----------
    base_estimator : object, default=None
        The base estimator to fit on random subsets of the dataset.
        If None, then the base estimator is a
        :class:`~sklearn.tree.DecisionTreeClassifier`.

    n_estimators : int, default=10
        The number of base estimators in the ensemble.

    max_samples : int or float, default=1.0
        The number of samples to draw from X to train each base estimator (with
        replacement by default, see `bootstrap` for more details).

        - If int, then draw `max_samples` samples.
        - If float, then draw `max_samples * X.shape[0]` samples.

    max_features : int or float, default=1.0
        The number of features to draw from X to train each base estimator (
        without replacement by default, see `bootstrap_features` for more
        details).

        - If int, then draw `max_features` features.
        - If float, then draw `max_features * X.shape[1]` features.

    bootstrap : bool, default=True
        Whether samples are drawn with replacement. If False, sampling
        without replacement is performed.

    bootstrap_features : bool, default=False
        Whether features are drawn with replacement.

    warm_start : bool, default=False
        When set to True, reuse the solution of the previous call to fit
        and add more estimators to the ensemble, otherwise, just fit
        a whole new ensemble. See :term:`the Glossary <warm_start>`.

    random_state : int, RandomState instance or None, default=None
        Controls the random resampling of the original dataset
        (sample wise and feature wise).
        If the base estimator accepts a `random_state` attribute, a different
        seed is generated for each instance in the ensemble.
        Pass an int for reproducible output across multiple function calls.
        See :term:`Glossary <random_state>`.

    Attributes
    ----------
    base_estimator_ : estimator
        The base estimator from which the ensemble is grown.

    estimators_ : list of estimators
        The collection of fitted base estimators.

    estimators_features_ : list of arrays
        The subset of drawn features for each base estimator.

    classes_ : ndarray of shape (n_classes,)
        The classes labels.

    n_classes_ : int or list
        The number of classes.

    See Also
    --------
    BaggingRegressor : A Bagging regressor.

    References
    ----------

    .. [1] L. Breiman, "Pasting small votes for classification in large
           databases and on-line", Machine Learning, 36(1), 85-103, 1999.

    .. [2] L. Breiman, "Bagging predictors", Machine Learning, 24(2), 123-140,
           1996.

    .. [3] T. Ho, "The random subspace method for constructing decision
           forests", Pattern Analysis and Machine Intelligence, 20(8), 832-844,
           1998.

    .. [4] G. Louppe and P. Geurts, "Ensembles on Random Patches", Machine
           Learning and Knowledge Discovery in Databases, 346-361, 2012.

    Examples
    --------
    >>> from sklearn.svm import SVC
    >>> from mars.learn.ensemble import BaggingClassifier
    >>> from mars.learn.datasets import make_classification
    >>> X, y = make_classification(n_samples=100, n_features=4,
    ...                            n_informative=2, n_redundant=0,
    ...                            random_state=0, shuffle=False)
    >>> clf = BaggingClassifier(base_estimator=SVC(),
    ...                         n_estimators=10, random_state=0).fit(X, y)
    >>> clf.predict([[0, 0, 0, 0]])
    array([1])
    """

    def _validate_y(self, y, session=None, run_kwargs=None):
        to_run = [check_classification_targets(y)]
        y = column_or_1d(y, warn=True)
        to_run.extend(mt.unique(y, return_inverse=True))
        _, self.classes_, y = execute(
            *to_run, session=session, **(run_kwargs or dict())
        )
        self.n_classes_ = len(self.classes_)

        return y

    def _predict_proba(self, X):
        check_is_fitted(self)
        X = convert_to_tensor_or_dataframe(X)
        predict_op = BaggingPredictionOperand(
            n_classes=self.n_classes_,
            prediction_type=PredictionType.PROBABILITY,
        )
        return predict_op(X, self.estimators_, self.estimator_features_)

    def predict(self, X, session=None, run_kwargs=None):
        """
        Predict class for X.

        The predicted class of an input sample is computed as the class with
        the highest mean predicted probability. If base estimators do not
        implement a ``predict_proba`` method, then it resorts to voting.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.

        Returns
        -------
        y : ndarray of shape (n_samples,)
            The predicted classes.
        """
        probas = self._predict_proba(X)
        y = self.classes_.take(mt.argmax(probas, axis=1), axis=0)
        return execute(y, session=session, **(run_kwargs or dict()))

    def predict_proba(self, X, session=None, run_kwargs=None):
        """
        Predict class probabilities for X.

        The predicted class probabilities of an input sample is computed as
        the mean predicted class probabilities of the base estimators in the
        ensemble. If base estimators do not implement a ``predict_proba``
        method, then it resorts to voting and the predicted class probabilities
        of an input sample represents the proportion of estimators predicting
        each class.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.

        Returns
        -------
        p : ndarray of shape (n_samples, n_classes)
            The class probabilities of the input samples. The order of the
            classes corresponds to that in the attribute :term:`classes_`.
        """
        probas = self._predict_proba(X)
        return execute(probas, session=session, **(run_kwargs or dict()))

    def predict_log_proba(self, X, session=None, run_kwargs=None):
        """
        Predict class log-probabilities for X.

        The predicted class log-probabilities of an input sample is computed as
        the log of the mean predicted class probabilities of the base
        estimators in the ensemble.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.

        Returns
        -------
        p : ndarray of shape (n_samples, n_classes)
            The class log-probabilities of the input samples. The order of the
            classes corresponds to that in the attribute :term:`classes_`.
        """
        check_is_fitted(self)
        X = convert_to_tensor_or_dataframe(X)
        predict_op = BaggingPredictionOperand(
            n_classes=self.n_classes_,
            prediction_type=PredictionType.LOG_PROBABILITY,
        )
        probas = predict_op(X, self.estimators_, self.estimator_features_)
        return execute(probas, session=session, **(run_kwargs or dict()))

    def decision_function(self, X, session=None, run_kwargs=None):
        """
        Average of the decision functions of the base classifiers.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.

        Returns
        -------
        score : ndarray of shape (n_samples, k)
            The decision function of the input samples. The columns correspond
            to the classes in sorted order, as they appear in the attribute
            ``classes_``. Regression and binary classification are special
            cases with ``k == 1``, otherwise ``k==n_classes``.
        """
        check_is_fitted(self)
        X = convert_to_tensor_or_dataframe(X)
        predict_op = BaggingPredictionOperand(
            n_classes=self.n_classes_,
            prediction_type=PredictionType.DECISION_FUNCTION,
        )
        result = predict_op(X, self.estimators_, self.estimator_features_)
        return execute(result, session=session, **(run_kwargs or dict()))


class BaggingRegressor(RegressorMixin, BaseBagging):
    """
    A Bagging regressor.

    A Bagging regressor is an ensemble meta-estimator that fits base
    regressors each on random subsets of the original dataset and then
    aggregate their individual predictions (either by voting or by averaging)
    to form a final prediction. Such a meta-estimator can typically be used as
    a way to reduce the variance of a black-box estimator (e.g., a decision
    tree), by introducing randomization into its construction procedure and
    then making an ensemble out of it.

    This algorithm encompasses several works from the literature. When random
    subsets of the dataset are drawn as random subsets of the samples, then
    this algorithm is known as Pasting [1]_. If samples are drawn with
    replacement, then the method is known as Bagging [2]_. When random subsets
    of the dataset are drawn as random subsets of the features, then the method
    is known as Random Subspaces [3]_. Finally, when base estimators are built
    on subsets of both samples and features, then the method is known as
    Random Patches [4]_.

    Read more in the :ref:`User Guide <bagging>`.

    Parameters
    ----------
    base_estimator : object, default=None
        The base estimator to fit on random subsets of the dataset.
        If None, then the base estimator is a
        :class:`~sklearn.tree.DecisionTreeRegressor`.

    n_estimators : int, default=10
        The number of base estimators in the ensemble.

    max_samples : int or float, default=1.0
        The number of samples to draw from X to train each base estimator (with
        replacement by default, see `bootstrap` for more details).

        - If int, then draw `max_samples` samples.
        - If float, then draw `max_samples * X.shape[0]` samples.

    max_features : int or float, default=1.0
        The number of features to draw from X to train each base estimator (
        without replacement by default, see `bootstrap_features` for more
        details).

        - If int, then draw `max_features` features.
        - If float, then draw `max_features * X.shape[1]` features.

    bootstrap : bool, default=True
        Whether samples are drawn with replacement. If False, sampling
        without replacement is performed.

    bootstrap_features : bool, default=False
        Whether features are drawn with replacement.

    warm_start : bool, default=False
        When set to True, reuse the solution of the previous call to fit
        and add more estimators to the ensemble, otherwise, just fit
        a whole new ensemble. See :term:`the Glossary <warm_start>`.

    random_state : int, RandomState instance or None, default=None
        Controls the random resampling of the original dataset
        (sample wise and feature wise).
        If the base estimator accepts a `random_state` attribute, a different
        seed is generated for each instance in the ensemble.
        Pass an int for reproducible output across multiple function calls.
        See :term:`Glossary <random_state>`.

    Attributes
    ----------
    base_estimator_ : estimator
        The base estimator from which the ensemble is grown.

    estimators_ : list of estimators
        The collection of fitted sub-estimators.

    estimators_features_ : list of arrays
        The subset of drawn features for each base estimator.

    See Also
    --------
    BaggingClassifier : A Bagging classifier.

    References
    ----------

    .. [1] L. Breiman, "Pasting small votes for classification in large
           databases and on-line", Machine Learning, 36(1), 85-103, 1999.

    .. [2] L. Breiman, "Bagging predictors", Machine Learning, 24(2), 123-140,
           1996.

    .. [3] T. Ho, "The random subspace method for constructing decision
           forests", Pattern Analysis and Machine Intelligence, 20(8), 832-844,
           1998.

    .. [4] G. Louppe and P. Geurts, "Ensembles on Random Patches", Machine
           Learning and Knowledge Discovery in Databases, 346-361, 2012.

    Examples
    --------
    >>> from sklearn.svm import SVR
    >>> from mars.learn.ensemble import BaggingRegressor
    >>> from mars.learn.datasets import make_regression
    >>> X, y = make_regression(n_samples=100, n_features=4,
    ...                        n_informative=2, n_targets=1,
    ...                        random_state=0, shuffle=False)
    >>> regr = BaggingRegressor(base_estimator=SVR(),
    ...                         n_estimators=10, random_state=0).fit(X, y)
    >>> regr.predict([[0, 0, 0, 0]])
    array([-2.8720...])
    """

    def predict(self, X, session=None, run_kwargs=None):
        """
        Predict regression target for X.

        The predicted regression target of an input sample is computed as the
        mean predicted regression targets of the estimators in the ensemble.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.

        Returns
        -------
        y : ndarray of shape (n_samples,)
            The predicted values.
        """
        check_is_fitted(self)
        X = convert_to_tensor_or_dataframe(X)
        predict_op = BaggingPredictionOperand(
            prediction_type=PredictionType.REGRESSION,
        )
        probas = predict_op(X, self.estimators_, self.estimator_features_)
        return execute(probas, session=session, **(run_kwargs or dict()))



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Dict, List, Union

from ..core import SessionInfo


class AbstractSessionAPI(ABC):
    @abstractmethod
    async def get_sessions(self) -> List[SessionInfo]:
        """
        Get information of all sessions

        Returns
        -------
        session_infos : List[SessionInfo]
            List of session infos.
        """

    @abstractmethod
    async def create_session(self, session_id: str) -> str:
        """
        Create session and return address.

        Parameters
        ----------
        session_id : str
            Session ID

        Returns
        -------
        address : str
            Session address.
        """

    @abstractmethod
    async def delete_session(self, session_id: str):
        """
        Delete session.

        Parameters
        ----------
        session_id : str
            Session ID.
        """

    @abstractmethod
    async def delete_all_sessions(self):
        """
        Delete all sessions.
        """

    @abstractmethod
    async def get_last_idle_time(
        self, session_id: Union[str, None] = None
    ) -> Union[float, None]:
        """
        Get session last idle time.

        Parameters
        ----------
        session_id : str, None
            Session ID. None for all sessions.

        Returns
        -------
        last_idle_time: str
            The last idle time if the session(s) is idle else None.
        """

    @abstractmethod
    async def fetch_tileable_op_logs(
        self,
        session_id: str,
        tileable_op_key: str,
        chunk_op_key_to_offsets: Dict[str, List[int]],
        chunk_op_key_to_sizes: Dict[str, List[int]],
    ) -> Dict:
        """
        Fetch tileable op's logs

        Parameters
        ----------
        session_id : str
            Session ID.
        tileable_op_key : str
            Tileable op key.
        chunk_op_key_to_offsets : str or int or list of int
            Fetch offsets.
        chunk_op_key_to_sizes : str or int or list of int
            Fetch sizes.

        Returns
        -------
        logs : dict
            chunk op key to result.
        """



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin

from ... import tensor as mt
from ...tensor.linalg import randomized_svd
from ...tensor.utils import check_random_state
from ...core import ExecutableTuple
from ..utils import check_array

__all__ = ["TruncatedSVD"]


class TruncatedSVD(BaseEstimator, TransformerMixin):
    """Dimensionality reduction using truncated SVD (aka LSA).

    This transformer performs linear dimensionality reduction by means of
    truncated singular value decomposition (SVD). Contrary to PCA, this
    estimator does not center the data before computing the singular value
    decomposition. This means it can work with scipy.sparse matrices
    efficiently.

    In particular, truncated SVD works on term count/tf-idf matrices as
    returned by the vectorizers in sklearn.feature_extraction.text. In that
    context, it is known as latent semantic analysis (LSA).

    This estimator supports two algorithms: a fast randomized SVD solver, and
    a "naive" algorithm that uses ARPACK as an eigensolver on (X * X.T) or
    (X.T * X), whichever is more efficient.

    Read more in the :ref:`User Guide <LSA>`.

    Parameters
    ----------
    n_components : int, default = 2
        Desired dimensionality of output data.
        Must be strictly less than the number of features.
        The default value is useful for visualisation. For LSA, a value of
        100 is recommended.

    algorithm : string, default = "randomized"
        SVD solver to use. Either "arpack" for the ARPACK wrapper in SciPy
        (scipy.sparse.linalg.svds), or "randomized" for the randomized
        algorithm due to Halko (2009).

    n_iter : int, optional (default 5)
        Number of iterations for randomized SVD solver. Not used by ARPACK.
        The default is larger than the default in `randomized_svd` to handle
        sparse matrices that may have large slowly decaying spectrum.

    random_state : int, RandomState instance or None, optional, default = None
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    tol : float, optional
        Tolerance for ARPACK. 0 means machine precision. Ignored by randomized
        SVD solver.

    Attributes
    ----------
    components_ : array, shape (n_components, n_features)

    explained_variance_ : array, shape (n_components,)
        The variance of the training samples transformed by a projection to
        each component.

    explained_variance_ratio_ : array, shape (n_components,)
        Percentage of variance explained by each of the selected components.

    singular_values_ : array, shape (n_components,)
        The singular values corresponding to each of the selected components.
        The singular values are equal to the 2-norms of the ``n_components``
        variables in the lower-dimensional space.

    Examples
    --------
    >>> from mars.learn.decomposition import TruncatedSVD
    >>> import mars.tensor as mt
    >>> from sklearn.random_projection import sparse_random_matrix
    >>> X = mt.tensor(sparse_random_matrix(100, 100, density=0.01, random_state=42))
    >>> svd = TruncatedSVD(n_components=5, n_iter=7, random_state=42)
    >>> svd.fit(X)  # doctest: +NORMALIZE_WHITESPACE
    TruncatedSVD(algorithm='randomized', n_components=5, n_iter=7,
            random_state=42, tol=0.0)
    >>> print(svd.explained_variance_ratio_)  # doctest: +ELLIPSIS
    [0.0606... 0.0584... 0.0497... 0.0434... 0.0372...]
    >>> print(svd.explained_variance_ratio_.sum())  # doctest: +ELLIPSIS
    0.249...
    >>> print(svd.singular_values_)  # doctest: +ELLIPSIS
    [2.5841... 2.5245... 2.3201... 2.1753... 2.0443...]

    See also
    --------
    PCA

    References
    ----------
    Finding structure with randomness: Stochastic algorithms for constructing
    approximate matrix decompositions
    Halko, et al., 2009 (arXiv:909) https://arxiv.org/pdf/0909.4061.pdf

    Notes
    -----
    SVD suffers from a problem called "sign indeterminacy", which means the
    sign of the ``components_`` and the output from transform depend on the
    algorithm and random state. To work around this, fit instances of this
    class to data once, then keep the instance around to do transformations.

    """

    def __init__(
        self,
        n_components=2,
        algorithm="randomized",
        n_iter=5,
        random_state=None,
        tol=0.0,
    ):
        self.algorithm = algorithm
        self.n_components = n_components
        self.n_iter = n_iter
        self.random_state = random_state
        self.tol = tol

    def fit(self, X, y=None, session=None):
        """Fit LSI model on training data X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Training data.
        session : session to run

        y : Ignored

        Returns
        -------
        self : object
            Returns the transformer object.
        """
        self.fit_transform(X, session=session)
        return self

    def fit_transform(self, X, y=None, session=None):
        """Fit LSI model to X and perform dimensionality reduction on X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Training data.
        session : session to run

        y : Ignored

        Returns
        -------
        X_new : array, shape (n_samples, n_components)
            Reduced version of X. This will always be a dense array.
        """
        X = check_array(X, accept_sparse=["csr", "csc"], ensure_min_features=2)
        random_state = check_random_state(self.random_state)

        if self.algorithm == "arpack":
            # U, Sigma, VT = svds(X, k=self.n_components, tol=self.tol)
            # # svds doesn't abide by scipy.linalg.svd/randomized_svd
            # # conventions, so reverse its outputs.
            # Sigma = Sigma[::-1]
            # U, VT = svd_flip(U[:, ::-1], VT[::-1])
            raise NotImplementedError("Does not support arpack for truncated_svd")

        elif self.algorithm == "randomized":
            k = self.n_components
            n_features = X.shape[1]
            if k >= n_features:
                raise ValueError(
                    f"n_components must be < n_features; got {k} >= {n_features}"
                )
            U, Sigma, VT = randomized_svd(
                X, self.n_components, n_iter=self.n_iter, random_state=random_state
            )
        else:
            raise ValueError(f"unknown algorithm {self.algorithm!r}")

        self.components_ = VT

        # Calculate explained variance & explained variance ratio
        X_transformed = U * Sigma
        self.explained_variance_ = exp_var = np.var(X_transformed, axis=0)
        full_var = mt.var(X, axis=0).sum()
        self.explained_variance_ratio_ = exp_var / full_var
        self.singular_values_ = Sigma  # Store the singular values.

        to_run_tensors = [
            X_transformed,
            self.components_,
            self.explained_variance_,
            self.explained_variance_ratio_,
            self.singular_values_,
        ]

        ExecutableTuple(to_run_tensors).execute(session=session)
        return X_transformed

    def transform(self, X, session=None):
        """Perform dimensionality reduction on X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            New data.
        session : session to run

        Returns
        -------
        X_new : array, shape (n_samples, n_components)
            Reduced version of X. This will always be a dense array.
        """
        X = check_array(X, accept_sparse="csr")
        ret = mt.dot(X, self.components_.T)
        ret.execute(session=session)
        return ret

    def inverse_transform(self, X, session=None):
        """Transform X back to its original space.

        Returns an array X_original whose transform would be X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_components)
            New data.
        session : session to run

        Returns
        -------
        X_original : array, shape (n_samples, n_features)
            Note that this is always a dense array.
        """
        X = check_array(X)
        ret = mt.dot(X, self.components_)
        ret.execute(session=session)
        return ret



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

try:
    import scipy.sparse as sps
except ImportError:  # pragma: no cover
    sps = None

from ... import opcodes
from ...core import ENTITY_TYPE, recursive_tile
from ...core.operand import OperandStage
from ...serialization.serializables import (
    KeyField,
    AnyField,
    StringField,
    Int64Field,
    BoolField,
)
from ...tensor import tensor as astensor
from ...utils import lazy_import, pd_release_version
from ..core import Index as DataFrameIndexType, INDEX_TYPE
from ..initializer import Index as asindex
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import validate_axis_style_args, parse_index
from .index_lib import DataFrameReindexHandler


cudf = lazy_import("cudf")

# under pandas<1.1, SparseArray ignores zeros on creation
_pd_sparse_miss_zero = pd_release_version[:2] < (1, 1)


class DataFrameReindex(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.REINDEX

    _input = KeyField("input")
    _index = AnyField("index")
    _index_freq = AnyField("index_freq")
    _columns = AnyField("columns")
    _method = StringField("method")
    _level = AnyField("level")
    _fill_value = AnyField("fill_value")
    _limit = Int64Field("limit")
    _enable_sparse = BoolField("enable_sparse")

    def __init__(
        self,
        index=None,
        index_freq=None,
        columns=None,
        method=None,
        level=None,
        fill_value=None,
        limit=None,
        enable_sparse=None,
        **kw,
    ):
        super().__init__(
            _index=index,
            _index_freq=index_freq,
            _columns=columns,
            _method=method,
            _level=level,
            _fill_value=fill_value,
            _limit=limit,
            _enable_sparse=enable_sparse,
            **kw,
        )

    @property
    def input(self):
        return self._input

    @property
    def index(self):
        return self._index

    @property
    def index_freq(self):
        return self._index_freq

    @property
    def columns(self):
        return self._columns

    @property
    def method(self):
        return self._method

    @property
    def level(self):
        return self._level

    @property
    def fill_value(self):
        return self._fill_value

    @property
    def limit(self):
        return self._limit

    @property
    def enable_sparse(self):
        return self._enable_sparse

    @property
    def _indexes(self):
        # used for index_lib
        indexes = []
        names = ("index", "columns")
        for ax in range(self.input.ndim):
            index = names[ax]
            val = getattr(self, index)
            if val is not None:
                indexes.append(val)
            else:
                indexes.append(slice(None))
        return indexes

    @_indexes.setter
    def _indexes(self, new_indexes):
        for index_field, new_index in zip(["_index", "_columns"], new_indexes):
            setattr(self, index_field, new_index)

    @property
    def indexes(self):
        return self._indexes

    @property
    def can_index_miss(self):
        return True

    def _new_chunks(self, inputs, kws=None, **kw):
        if self.stage == OperandStage.map and len(inputs) < len(self._inputs):
            assert len(inputs) == len(self._inputs) - 1
            inputs.append(self._fill_value.chunks[0])

        if self.stage == OperandStage.agg and self._fill_value is not None:
            # fill_value is not required
            self._fill_value = None

        return super()._new_chunks(inputs, kws=kws, **kw)

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        self._input = next(inputs_iter)
        if self._index is not None and isinstance(self._index, ENTITY_TYPE):
            self._index = next(inputs_iter)
        if self._fill_value is not None and isinstance(self._fill_value, ENTITY_TYPE):
            self._fill_value = next(inputs_iter)

    def __call__(self, df_or_series):
        inputs = [df_or_series]
        shape = list(df_or_series.shape)
        index_value = df_or_series.index_value
        columns_value = dtypes = None
        if df_or_series.ndim == 2:
            columns_value = df_or_series.columns_value
            dtypes = df_or_series.dtypes

        if self._index is not None:
            shape[0] = self._index.shape[0]
            index_value = asindex(self._index).index_value
            self._index = astensor(self._index)
            if isinstance(self._index, ENTITY_TYPE):
                inputs.append(self._index)
        if self._columns is not None:
            shape[1] = self._columns.shape[0]
            dtypes = df_or_series.dtypes.reindex(index=self._columns).fillna(
                np.dtype(np.float64)
            )
            columns_value = parse_index(dtypes.index, store_data=True)
        if self._fill_value is not None and isinstance(self._fill_value, ENTITY_TYPE):
            inputs.append(self._fill_value)

        if df_or_series.ndim == 1:
            return self.new_series(
                inputs,
                shape=tuple(shape),
                dtype=df_or_series.dtype,
                index_value=index_value,
                name=df_or_series.name,
            )
        else:
            return self.new_dataframe(
                inputs,
                shape=tuple(shape),
                dtypes=dtypes,
                index_value=index_value,
                columns_value=columns_value,
            )

    @classmethod
    def tile(cls, op):
        if all(len(inp.chunks) == 1 for inp in op.inputs):
            # tile one chunk
            out = op.outputs[0]

            chunk_op = op.copy().reset_key()
            chunk_params = out.params.copy()
            chunk_params["index"] = (0,) * out.ndim
            out_chunk = chunk_op.new_chunk(
                [inp.chunks[0] for inp in op.inputs], kws=[chunk_params]
            )

            params = out.params.copy()
            params["nsplits"] = ((s,) for s in out.shape)
            params["chunks"] = [out_chunk]
            new_op = op.copy()
            return new_op.new_tileables(op.inputs, kws=[params])

        handler = DataFrameReindexHandler()
        result = yield from handler.handle(op)
        if op.method is None and op.fill_value is None:
            return [result]
        else:
            axis = 1 if op.columns is not None and op.index is None else 0
            result = result.fillna(
                value=op.fill_value, method=op.method, axis=axis, limit=op.limit
            )
            return [(yield from recursive_tile(result))]

    @classmethod
    def _get_value(cls, ctx, obj):
        if obj is not None and hasattr(obj, "key"):
            return ctx[obj.key]
        return obj

    @classmethod
    def _convert_to_writable(cls, obj):
        if isinstance(obj, np.ndarray) and not obj.flags.writeable:
            return obj.copy()
        return obj

    @classmethod
    def _sparse_reindex(cls, inp, index=None, columns=None):
        if inp.ndim == 2:
            columns = inp.columns if columns is None else columns
            index_shape = len(index) if index is not None else len(inp)
            i_to_columns = dict()

            for i, col in enumerate(columns):
                if col in inp.dtypes:
                    if index is None:
                        i_to_columns[i] = inp[col]
                    else:
                        indexer = inp.index.reindex(index)[1]
                        cond = indexer >= 0
                        available_indexer = indexer[cond]
                        del indexer
                        data = inp[col].iloc[available_indexer].to_numpy()
                        ind = cond.nonzero()[0]
                        spmatrix = sps.csc_matrix(
                            (data, (ind, np.zeros_like(ind))),
                            shape=(index_shape, 1),
                            dtype=inp[col].dtype,
                        )
                        # convert to SparseDtype(xxx, np.nan)
                        # to ensure 0 in sparse_array not converted to np.nan
                        if not _pd_sparse_miss_zero:
                            sparse_array = pd.arrays.SparseArray.from_spmatrix(spmatrix)
                            sparse_array = pd.arrays.SparseArray(
                                sparse_array.sp_values,
                                sparse_index=sparse_array.sp_index,
                                fill_value=np.nan,
                                dtype=pd.SparseDtype(sparse_array.dtype, np.nan),
                            )
                        else:
                            from pandas._libs.sparse import IntIndex

                            sparse_array = pd.arrays.SparseArray(
                                data,
                                sparse_index=IntIndex(index_shape, ind),
                                fill_value=np.nan,
                                dtype=pd.SparseDtype(data.dtype, np.nan),
                            )
                        series = pd.Series(sparse_array, index=index)

                        i_to_columns[i] = series
                else:
                    ind = index if index is not None else inp.index
                    i_to_columns[i] = pd.DataFrame.sparse.from_spmatrix(
                        sps.coo_matrix((index_shape, 1), dtype=np.float64), index=ind
                    ).iloc[:, 0]

            df = pd.DataFrame(i_to_columns)
            df.columns = columns
            return df
        else:
            indexer = inp.index.reindex(index)[1]
            cond = indexer >= 0
            available_indexer = indexer[cond]
            del indexer
            data = inp.iloc[available_indexer].to_numpy()
            ind = cond.nonzero()[0]
            spmatrix = sps.csc_matrix(
                (data, (ind, np.zeros_like(ind))),
                shape=(len(index), 1),
                dtype=inp.dtype,
            )
            sparse_array = pd.arrays.SparseArray.from_spmatrix(spmatrix)
            # convert to SparseDtype(xxx, np.nan)
            # to ensure 0 in sparse_array not converted to np.nan
            sparse_array = pd.arrays.SparseArray(
                sparse_array.sp_values,
                sparse_index=sparse_array.sp_index,
                fill_value=np.nan,
                dtype=pd.SparseDtype(sparse_array.dtype, np.nan),
            )
            series = pd.Series(sparse_array, index=index, name=inp.name)
            return series

    @classmethod
    def _reindex(cls, ctx, op, fill=True, try_sparse=None):
        inp = cls._convert_to_writable(ctx[op.input.key])
        index = cls._get_value(ctx, op.index)
        if op.index_freq is not None:
            index = pd.Index(index, freq=op.index_freq)
        columns = cls._get_value(ctx, op.columns)
        kw = {"level": op.level}
        if index is not None and not isinstance(index, slice):
            kw["index"] = cls._convert_to_writable(index)
        if columns is not None and not isinstance(columns, slice):
            kw["columns"] = cls._convert_to_writable(columns)
        if fill:
            kw["method"] = op.method
            kw["fill_value"] = cls._get_value(ctx, op.fill_value)
            kw["limit"] = op.limit

        if (
            try_sparse
            and not fill
            and op.level is None
            and isinstance(inp, (pd.DataFrame, pd.Series))
            and sps is not None
        ):
            # 1. sparse is used in map only
            # 2. for MultiIndex, sparse is not needed as well
            # 3. only consider cpu
            # 4. scipy is installed

            if op.enable_sparse is None:
                # try to use sparse if estimated size > 2 * input_size
                cur_size = inp.memory_usage(deep=True)
                if inp.ndim == 2:
                    cur_size = cur_size.sum()
                element_size = cur_size / inp.size
                shape = list(inp.shape)
                if "index" in kw:
                    shape[0] = len(kw["index"])
                if "columns" in kw:
                    shape[1] = len(kw["columns"])
                estimate_size = np.prod(shape) * element_size

                fitted = estimate_size > cur_size * 2
            else:
                # specified when op.enable_sparse == True
                fitted = True

            if fitted:
                # use sparse instead
                return cls._sparse_reindex(
                    inp, index=kw.get("index"), columns=kw.get("columns")
                )

        return inp.reindex(**kw)

    @classmethod
    def _execute_reindex(cls, ctx, op):
        ctx[op.outputs[0].key] = cls._reindex(ctx, op)

    @classmethod
    def _execute_map(cls, ctx, op):
        if op.enable_sparse is not None:
            try_sparse = op.enable_sparse
        else:
            try_sparse = True
        ctx[op.outputs[0].key] = cls._reindex(
            ctx, op, fill=False, try_sparse=try_sparse
        )

    @classmethod
    def _convert_to_dense(cls, series):
        if isinstance(series.dtype, pd.SparseDtype):
            return series.astype(
                pd.SparseDtype(series.dtype.subtype, np.nan)
            ).sparse.to_dense()
        return series

    @classmethod
    def _merge_chunks(cls, inputs):
        xdf = cls._get_xdf(inputs[0])

        ndim = inputs[0].ndim
        if ndim == 2:
            columns = inputs[0].columns
            result = xdf.DataFrame(
                np.full(inputs[0].shape, np.nan), columns=columns, index=inputs[0].index
            )
        else:
            columns = [inputs[0].name]
            result = None

        for i in range(len(columns)):
            if ndim == 1:
                curr = cls._convert_to_dense(inputs[0]).copy()
            else:
                curr = cls._convert_to_dense(inputs[0].iloc[:, i]).copy()
            for j in range(len(inputs) - 1):
                if ndim == 2:
                    left = cls._convert_to_dense(inputs[j].iloc[:, i])
                    right = cls._convert_to_dense(inputs[j + 1].iloc[:, i])
                else:
                    left = cls._convert_to_dense(inputs[j])
                    right = cls._convert_to_dense(inputs[j + 1])

                left_notna = left.notna()
                right_notna = right.notna()
                if (left_notna & right_notna).sum() > 0:
                    raise ValueError("cannot reindex from a duplicate axis")
                curr.loc[left_notna] = left.loc[left_notna]
                curr.loc[right_notna] = right.loc[right_notna]
            if ndim == 1:
                result = curr
            else:
                result.iloc[:, i] = curr

        return result

    @classmethod
    def _get_xdf(cls, obj):
        return (
            pd if isinstance(obj, (pd.DataFrame, pd.Series)) or cudf is None else cudf
        )

    @classmethod
    def _execute_agg(cls, ctx, op):
        out = op.outputs[0]

        if op.index is None and op.columns is None:
            # index is tensor
            inputs = [ctx[inp.key] for inp in op.inputs]

            xdf = cls._get_xdf(inputs[0])

            if inputs[0].index.nlevels > 1 and op.level is not None:
                # multi index
                result = xdf.concat(inputs)
            else:
                result = cls._merge_chunks(inputs) if len(inputs) > 1 else inputs[0]

            ctx[out.key] = result

        else:
            # ndarray index or columns
            if isinstance(op.index, slice) and op.index == slice(None):
                axis = 1
                labels = op.columns
            else:
                assert op.columns is None or (
                    isinstance(op.columns, slice) and op.columns == slice(None)
                )
                axis = 0
                labels = op.index

            inp = ctx[op.inputs[0].key]
            if inp.index.nlevels > 1 and op.level is not None:
                new_inp = inp
            else:
                # split input
                size = out.shape[axis]
                assert inp.shape[axis] % size == 0
                inputs = []
                for i in range(inp.shape[axis] // size):
                    slc = [slice(None)] * inp.ndim
                    slc[axis] = slice(size * i, size * (i + 1))
                    inputs.append(inp.iloc[tuple(slc)])
                new_inp = cls._merge_chunks(inputs)

            labels = cls._convert_to_writable(labels)
            if out.ndim == 2:
                result = new_inp.reindex(labels=labels, axis=axis, level=op.level)
            else:
                result = new_inp.reindex(index=labels, level=op.level)
            ctx[out.key] = result

    @classmethod
    def execute(cls, ctx, op):
        if op.stage == OperandStage.map:
            return cls._execute_map(ctx, op)
        elif op.stage == OperandStage.agg:
            return cls._execute_agg(ctx, op)
        else:
            assert op.stage is None
            return cls._execute_reindex(ctx, op)


def reindex(df_or_series, *args, **kwargs):
    """
    Conform Series/DataFrame to new index with optional filling logic.

    Places NA/NaN in locations having no value in the previous index. A new object
    is produced unless the new index is equivalent to the current one and
    ``copy=False``.

    Parameters
    ----------
    labels : array-like, optional
        New labels / index to conform the axis specified by 'axis' to.
    index, columns : array-like, optional
        New labels / index to conform to, should be specified using
        keywords. Preferably an Index object to avoid duplicating data.
    axis : int or str, optional
        Axis to target. Can be either the axis name ('index', 'columns')
        or number (0, 1).
    method : {None, 'backfill'/'bfill', 'pad'/'ffill', 'nearest'}
        Method to use for filling holes in reindexed DataFrame.
        Please note: this is only applicable to DataFrames/Series with a
        monotonically increasing/decreasing index.

        * None (default): don't fill gaps
        * pad / ffill: Propagate last valid observation forward to next
          valid.
        * backfill / bfill: Use next valid observation to fill gap.
        * nearest: Use nearest valid observations to fill gap.

    copy : bool, default True
        Return a new object, even if the passed indexes are the same.
    level : int or name
        Broadcast across a level, matching Index values on the
        passed MultiIndex level.
    fill_value : scalar, default np.NaN
        Value to use for missing values. Defaults to NaN, but can be any
        "compatible" value.
    limit : int, default None
        Maximum number of consecutive elements to forward or backward fill.
    tolerance : optional
        Maximum distance between original and new labels for inexact
        matches. The values of the index at the matching locations most
        satisfy the equation ``abs(index[indexer] - target) <= tolerance``.

        Tolerance may be a scalar value, which applies the same tolerance
        to all values, or list-like, which applies variable tolerance per
        element. List-like includes list, tuple, array, Series, and must be
        the same size as the index and its dtype must exactly match the
        index's type.

    Returns
    -------
    Series/DataFrame with changed index.

    See Also
    --------
    DataFrame.set_index : Set row labels.
    DataFrame.reset_index : Remove row labels or move them to new columns.
    DataFrame.reindex_like : Change to same indices as other DataFrame.

    Examples
    --------

    ``DataFrame.reindex`` supports two calling conventions

    * ``(index=index_labels, columns=column_labels, ...)``
    * ``(labels, axis={'index', 'columns'}, ...)``

    We *highly* recommend using keyword arguments to clarify your
    intent.

    Create a dataframe with some fictional data.

    >>> import mars.dataframe as md
    >>> index = ['Firefox', 'Chrome', 'Safari', 'IE10', 'Konqueror']
    >>> df = md.DataFrame({'http_status': [200, 200, 404, 404, 301],
    ...                   'response_time': [0.04, 0.02, 0.07, 0.08, 1.0]},
    ...                   index=index)
    >>> df.execute()
               http_status  response_time
    Firefox            200           0.04
    Chrome             200           0.02
    Safari             404           0.07
    IE10               404           0.08
    Konqueror          301           1.00

    Create a new index and reindex the dataframe. By default
    values in the new index that do not have corresponding
    records in the dataframe are assigned ``NaN``.

    >>> new_index = ['Safari', 'Iceweasel', 'Comodo Dragon', 'IE10',
    ...              'Chrome']
    >>> df.reindex(new_index).execute()
                   http_status  response_time
    Safari               404.0           0.07
    Iceweasel              NaN            NaN
    Comodo Dragon          NaN            NaN
    IE10                 404.0           0.08
    Chrome               200.0           0.02

    We can fill in the missing values by passing a value to
    the keyword ``fill_value``. Because the index is not monotonically
    increasing or decreasing, we cannot use arguments to the keyword
    ``method`` to fill the ``NaN`` values.

    >>> df.reindex(new_index, fill_value=0).execute()
                   http_status  response_time
    Safari                 404           0.07
    Iceweasel                0           0.00
    Comodo Dragon            0           0.00
    IE10                   404           0.08
    Chrome                 200           0.02

    >>> df.reindex(new_index, fill_value='missing').execute()
                  http_status response_time
    Safari                404          0.07
    Iceweasel         missing       missing
    Comodo Dragon     missing       missing
    IE10                  404          0.08
    Chrome                200          0.02

    We can also reindex the columns.

    >>> df.reindex(columns=['http_status', 'user_agent']).execute()
               http_status  user_agent
    Firefox            200         NaN
    Chrome             200         NaN
    Safari             404         NaN
    IE10               404         NaN
    Konqueror          301         NaN

    Or we can use "axis-style" keyword arguments

    >>> df.reindex(['http_status', 'user_agent'], axis="columns").execute()
               http_status  user_agent
    Firefox            200         NaN
    Chrome             200         NaN
    Safari             404         NaN
    IE10               404         NaN
    Konqueror          301         NaN

    To further illustrate the filling functionality in
    ``reindex``, we will create a dataframe with a
    monotonically increasing index (for example, a sequence
    of dates).

    >>> date_index = md.date_range('1/1/2010', periods=6, freq='D')
    >>> df2 = md.DataFrame({"prices": [100, 101, np.nan, 100, 89, 88]},
    ...                    index=date_index)
    >>> df2.execute()
                prices
    2010-01-01   100.0
    2010-01-02   101.0
    2010-01-03     NaN
    2010-01-04   100.0
    2010-01-05    89.0
    2010-01-06    88.0

    Suppose we decide to expand the dataframe to cover a wider
    date range.

    >>> date_index2 = md.date_range('12/29/2009', periods=10, freq='D')
    >>> df2.reindex(date_index2).execute()
                prices
    2009-12-29     NaN
    2009-12-30     NaN
    2009-12-31     NaN
    2010-01-01   100.0
    2010-01-02   101.0
    2010-01-03     NaN
    2010-01-04   100.0
    2010-01-05    89.0
    2010-01-06    88.0
    2010-01-07     NaN

    The index entries that did not have a value in the original data frame
    (for example, '2009-12-29') are by default filled with ``NaN``.
    If desired, we can fill in the missing values using one of several
    options.

    For example, to back-propagate the last valid value to fill the ``NaN``
    values, pass ``bfill`` as an argument to the ``method`` keyword.

    >>> df2.reindex(date_index2, method='bfill').execute()
                prices
    2009-12-29   100.0
    2009-12-30   100.0
    2009-12-31   100.0
    2010-01-01   100.0
    2010-01-02   101.0
    2010-01-03     NaN
    2010-01-04   100.0
    2010-01-05    89.0
    2010-01-06    88.0
    2010-01-07     NaN

    Please note that the ``NaN`` value present in the original dataframe
    (at index value 2010-01-03) will not be filled by any of the
    value propagation schemes. This is because filling while reindexing
    does not look at dataframe values, but only compares the original and
    desired indexes. If you do want to fill in the ``NaN`` values present
    in the original dataframe, use the ``fillna()`` method.

    See the :ref:`user guide <basics.reindexing>` for more.
    """
    axes = validate_axis_style_args(df_or_series, args, kwargs, "labels", "reindex")
    # Pop these, since the values are in `kwargs` under different names
    kwargs.pop("index", None)
    if df_or_series.ndim > 1:
        kwargs.pop("columns", None)
        kwargs.pop("axis", None)
        kwargs.pop("labels", None)
    method = kwargs.pop("method", None)
    level = kwargs.pop("level", None)
    copy = kwargs.pop("copy", True)
    limit = kwargs.pop("limit", None)
    tolerance = kwargs.pop("tolerance", None)
    fill_value = kwargs.pop("fill_value", None)
    enable_sparse = kwargs.pop("enable_sparse", None)

    if kwargs:
        raise TypeError(
            "reindex() got an unexpected keyword "
            f'argument "{list(kwargs.keys())[0]}"'
        )

    if tolerance is not None:  # pragma: no cover
        raise NotImplementedError("`tolerance` is not supported yet")

    if method == "nearest":  # pragma: no cover
        raise NotImplementedError("method=nearest is not supported yet")

    index = axes.get("index")
    index_freq = None
    if isinstance(index, ENTITY_TYPE):
        if isinstance(index, DataFrameIndexType):
            index_freq = getattr(index.index_value.value, "freq", None)
        if not isinstance(index, INDEX_TYPE):
            index = astensor(index)
    elif index is not None:
        index = np.asarray(index)
        index_freq = getattr(index, "freq", None)

    columns = axes.get("columns")
    if isinstance(columns, ENTITY_TYPE):  # pragma: no cover
        try:
            columns = columns.fetch()
        except ValueError:
            raise NotImplementedError(
                "`columns` need to be executed first if it's a Mars object"
            )
    elif columns is not None:
        columns = np.asarray(columns)

    if isinstance(fill_value, ENTITY_TYPE) and getattr(fill_value, "ndim", 0) != 0:
        raise ValueError("fill_value must be a scalar")

    op = DataFrameReindex(
        index=index,
        index_freq=index_freq,
        columns=columns,
        method=method,
        level=level,
        fill_value=fill_value,
        limit=limit,
        enable_sparse=enable_sparse,
    )
    ret = op(df_or_series)

    if copy:
        return ret.copy()
    return ret


def reindex_like(
    df_or_series, other, method=None, copy=True, limit=None, tolerance=None
):
    """
    Return an object with matching indices as other object.

    Conform the object to the same index on all axes. Optional
    filling logic, placing NaN in locations having no value
    in the previous index. A new object is produced unless the
    new index is equivalent to the current one and copy=False.

    Parameters
    ----------
    other : Object of the same data type
        Its row and column indices are used to define the new indices
        of this object.
    method : {None, 'backfill'/'bfill', 'pad'/'ffill', 'nearest'}
        Method to use for filling holes in reindexed DataFrame.
        Please note: this is only applicable to DataFrames/Series with a
        monotonically increasing/decreasing index.

        * None (default): don't fill gaps
        * pad / ffill: propagate last valid observation forward to next
          valid
        * backfill / bfill: use next valid observation to fill gap
        * nearest: use nearest valid observations to fill gap.

    copy : bool, default True
        Return a new object, even if the passed indexes are the same.
    limit : int, default None
        Maximum number of consecutive labels to fill for inexact matches.
    tolerance : optional
        Maximum distance between original and new labels for inexact
        matches. The values of the index at the matching locations must
        satisfy the equation ``abs(index[indexer] - target) <= tolerance``.

        Tolerance may be a scalar value, which applies the same tolerance
        to all values, or list-like, which applies variable tolerance per
        element. List-like includes list, tuple, array, Series, and must be
        the same size as the index and its dtype must exactly match the
        index's type.

    Returns
    -------
    Series or DataFrame
        Same type as caller, but with changed indices on each axis.

    See Also
    --------
    DataFrame.set_index : Set row labels.
    DataFrame.reset_index : Remove row labels or move them to new columns.
    DataFrame.reindex : Change to new indices or expand indices.

    Notes
    -----
    Same as calling
    ``.reindex(index=other.index, columns=other.columns,...)``.

    Examples
    --------
    >>> import pandas as pd
    >>> import mars.dataframe as md
    >>> df1 = md.DataFrame([[24.3, 75.7, 'high'],
    ...                     [31, 87.8, 'high'],
    ...                     [22, 71.6, 'medium'],
    ...                     [35, 95, 'medium']],
    ...                    columns=['temp_celsius', 'temp_fahrenheit',
    ...                             'windspeed'],
    ...                    index=md.date_range(start='2014-02-12',
    ...                                        end='2014-02-15', freq='D'))

    >>> df1.execute()
               temp_celsius temp_fahrenheit windspeed
    2014-02-12         24.3            75.7      high
    2014-02-13           31            87.8      high
    2014-02-14           22            71.6    medium
    2014-02-15           35              95    medium

    >>> df2 = md.DataFrame([[28, 'low'],
    ...                     [30, 'low'],
    ...                     [35.1, 'medium']],
    ...                    columns=['temp_celsius', 'windspeed'],
    ...                    index=pd.DatetimeIndex(['2014-02-12', '2014-02-13',
    ...                                            '2014-02-15']))

    >>> df2.execute()
                temp_celsius windspeed
    2014-02-12          28.0       low
    2014-02-13          30.0       low
    2014-02-15          35.1    medium

    >>> df2.reindex_like(df1).execute()
                temp_celsius  temp_fahrenheit windspeed
    2014-02-12          28.0              NaN       low
    2014-02-13          30.0              NaN       low
    2014-02-14           NaN              NaN       NaN
    2014-02-15          35.1              NaN    medium
    """
    cond = df_or_series.index_value.key == other.index_value.key
    if df_or_series.ndim == 2:
        cond &= df_or_series.columns_value.key == other.columns_value.key
    if cond and not copy:
        return df_or_series

    kw = {
        "index": other.index,
        "method": method,
        "limit": limit,
        "tolerance": tolerance,
    }
    if df_or_series.ndim == 2:
        kw["columns"] = other.dtypes.index
    return reindex(df_or_series, **kw)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...core import ExecutableTuple, recursive_tile
from ...lib.sparse.core import get_array_module
from ...serialization.serializables import KeyField, AnyField, Int32Field
from ..core import Tensor
from ..datasource import tensor as astensor
from ..utils import calc_sliced_size
from ..operands import TensorHasInput, TensorOperandMixin


class TensorSplit(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.ARRAY_SPLIT

    _input = KeyField("input")
    _indices_or_sections = AnyField("indices_or_sections")
    _axis = Int32Field("axis")

    def __init__(self, axis=None, **kw):
        super().__init__(_axis=axis, **kw)

    @property
    def indices_or_sections(self):
        return self._indices_or_sections

    @property
    def axis(self):
        return getattr(self, "_axis", 0)

    @property
    def output_limit(self):
        return float("inf")

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]
        if len(self._inputs) > 1:
            self._indices_or_sections = self._inputs[1]

    def __call__(self, a, indices_or_sections, is_split=False):
        axis = self._axis
        size = a.shape[axis]
        if np.isnan(size):
            raise ValueError(
                "cannot split array with unknown shape, "
                "call `.execute()` on input tensor first"
            )

        if (
            isinstance(indices_or_sections, Tensor)
            and hasattr(indices_or_sections.op, "data")
            and indices_or_sections.op.data is not None
        ):
            indices_or_sections = indices_or_sections.op.data

        try:
            indices_or_sections = int(indices_or_sections)
            if is_split:
                if size % indices_or_sections:
                    raise ValueError(
                        "tensor split does not result in an equal division"
                    )
                nparts = indices_or_sections
                nsplit = (size // indices_or_sections,) * nparts
            else:
                nparts = indices_or_sections
                if size % indices_or_sections == 0:
                    nsplit = (size // indices_or_sections,) * nparts
                else:
                    nsplit = (size // indices_or_sections + 1,) * (
                        size % indices_or_sections
                    ) + (size // indices_or_sections,) * (
                        size - size % indices_or_sections
                    )
        except TypeError:
            if isinstance(indices_or_sections, Tensor):
                nparts = indices_or_sections.shape[0] + 1
                nsplit = (np.nan,) * nparts
            else:
                ind = indices_or_sections = get_array_module(
                    indices_or_sections
                ).asarray(indices_or_sections)
                if indices_or_sections.ndim != 1 or not np.issubdtype(
                    indices_or_sections.dtype, np.integer
                ):
                    raise TypeError("slice indices must be integers or None")
                nparts = indices_or_sections.shape[0] + 1
                get = lambda i: None if i < 0 or i >= len(ind) else ind[i]
                nsplit = [
                    calc_sliced_size(size, slice(get(j - 1), get(j)))
                    for j in range(nparts)
                ]

        inputs = [a]
        if isinstance(indices_or_sections, Tensor):
            inputs.append(indices_or_sections)
        else:
            self._indices_or_sections = indices_or_sections

        kws = [
            {
                "i": i,
                "shape": a.shape[:axis] + (nsplit[i],) + a.shape[axis + 1 :],
                "order": a.order,
            }
            for i in range(nparts)
        ]
        return ExecutableTuple(self.new_tensors(inputs, kws=kws, output_limit=nparts))

    @classmethod
    def tile(cls, op):
        in_tensor = op.input
        splits = op.outputs
        axis = op.axis

        acc_shapes = np.cumsum([s.shape[axis] for s in splits])
        out_kws = [dict() for _ in splits]
        for i, split in enumerate(splits):
            slc = slice(0 if i == 0 else acc_shapes[i - 1], acc_shapes[i])
            new_s = yield from recursive_tile(in_tensor[(slice(None),) * axis + (slc,)])
            out_kws[i]["chunks"] = new_s.chunks
            out_kws[i]["nsplits"] = new_s.nsplits
            out_kws[i]["shape"] = split.shape
            out_kws[i]["order"] = op.outputs[i].order

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, kws=out_kws, output_limit=len(out_kws))


def _split(a, indices_or_sections, axis=0, is_split=False):
    op = TensorSplit(axis=axis, dtype=a.dtype)
    return op(a, indices_or_sections, is_split=is_split)


def split(ary, indices_or_sections, axis=0):
    """
    Split a tensor into multiple sub-tensors.

    Parameters
    ----------
    ary : Tensor
        Tensor to be divided into sub-tensors.
    indices_or_sections : int or 1-D tensor
        If `indices_or_sections` is an integer, N, the array will be divided
        into N equal tensors along `axis`.  If such a split is not possible,
        an error is raised.

        If `indices_or_sections` is a 1-D tensor of sorted integers, the entries
        indicate where along `axis` the array is split.  For example,
        ``[2, 3]`` would, for ``axis=0``, result in

          - ary[:2]
          - ary[2:3]
          - ary[3:]

        If an index exceeds the dimension of the tensor along `axis`,
        an empty sub-tensor is returned correspondingly.
    axis : int, optional
        The axis along which to split, default is 0.

    Returns
    -------
    sub-tensors : list of Tensors
        A list of sub-tensors.

    Raises
    ------
    ValueError
        If `indices_or_sections` is given as an integer, but
        a split does not result in equal division.

    See Also
    --------
    array_split : Split a tensor into multiple sub-tensors of equal or
                  near-equal size.  Does not raise an exception if
                  an equal division cannot be made.
    hsplit : Split  into multiple sub-arrays horizontally (column-wise).
    vsplit : Split tensor into multiple sub-tensors vertically (row wise).
    dsplit : Split tensor into multiple sub-tensors along the 3rd axis (depth).
    concatenate : Join a sequence of tensors along an existing axis.
    stack : Join a sequence of tensors along a new axis.
    hstack : Stack tensors in sequence horizontally (column wise).
    vstack : Stack tensors in sequence vertically (row wise).
    dstack : Stack tensors in sequence depth wise (along third dimension).

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.arange(9.0)
    >>> mt.split(x, 3).execute()
    [array([ 0.,  1.,  2.]), array([ 3.,  4.,  5.]), array([ 6.,  7.,  8.])]

    >>> x = mt.arange(8.0)
    >>> mt.split(x, [3, 5, 6, 10]).execute()
    [array([ 0.,  1.,  2.]),
     array([ 3.,  4.]),
     array([ 5.]),
     array([ 6.,  7.]),
     array([], dtype=float64)]

    """
    return _split(astensor(ary), indices_or_sections, axis=axis, is_split=True)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ...lib.filesystem import get_fs, FSMap
from ..core import TensorOrder
from .core import TensorFromHDF5Like


class TensorFromZarr(TensorFromHDF5Like):
    _op_type_ = OperandDef.TENSOR_FROM_ZARR

    @classmethod
    def execute(cls, ctx, op):
        import zarr

        axis_offsets = op.axis_offsets
        shape = op.outputs[0].shape

        fs = get_fs(op.filename, None)
        fs_map = FSMap(op.filename, fs)

        root = zarr.group(store=fs_map)
        path = cls.get_path(op.group, op.dataset)
        arr = root[path]

        data = arr[
            tuple(
                slice(offset, offset + size)
                for offset, size in zip(axis_offsets, shape)
            )
        ]
        ctx[op.outputs[0].key] = data


def fromzarr(path, group=None, dataset=None, chunk_size=None):
    import zarr

    try:
        # since v2.11.0, zarr convert mutable mappings to KVStore
        from zarr.storage import KVStore as zarr_kvstore
    except ImportError:  # pragma: no cover
        zarr_kvstore = None

    if isinstance(path, zarr.Array):
        arr = path
        if zarr_kvstore is None and isinstance(arr.store, FSMap):  # pragma: no cover
            root = arr.store.root
            path, dataset = root.rsplit("/", 1)
        elif zarr_kvstore and isinstance(arr.store, zarr_kvstore):
            root = arr.store._mutable_mapping.root
            path, dataset = root.rsplit("/", 1)
        else:
            path = arr.store.path
            if "/" in arr.path and group is None:
                group = arr.path.rsplit("/", 1)[0]
            dataset = arr.basename
            if not dataset:
                path, dataset = path.rsplit("/", 1)
        shape = arr.shape
    elif isinstance(path, str):
        fs = get_fs(path, None)
        fs_map = FSMap(path, fs)

        if group is None and dataset is None:
            arr = zarr.open(fs_map)
            if isinstance(arr, zarr.Array):
                return fromzarr(arr, chunk_size=chunk_size)

        g = zarr.group(store=fs_map)
        arr = g[TensorFromZarr.get_path(group, dataset)]
        shape = arr.shape
    else:
        raise TypeError(
            "`path` passed has wrong type, "
            "expect str, or zarr.Array"
            f"got {type(path)}"
        )

    chunk_size = chunk_size if chunk_size is not None else arr.chunks
    op = TensorFromZarr(filename=path, group=group, dataset=dataset, dtype=arr.dtype)
    return op(shape, chunk_size=chunk_size, order=TensorOrder(arr.order))
